#
# Copyright 2017 Ericsson India Global Services Pvt Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import abc
from unittest import mock
import urllib

from oslotest import base

from ceilometer import service
from networking_odl.ceilometer.network.statistics.opendaylight_v2 import driver
from oslo_utils import uuidutils

ADMIN_ID = str(uuidutils.generate_uuid())
PORT_1_TENANT_ID = str(uuidutils.generate_uuid())
PORT_2_TENANT_ID = str(uuidutils.generate_uuid())
PORT_1_ID = str(uuidutils.generate_uuid())
PORT_2_ID = str(uuidutils.generate_uuid())


class _Base(base.BaseTestCase, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def switch_data(self):
        pass

    fake_odl_url = urllib.parse.ParseResult('opendaylight.v2',
                                            'localhost:8080',
                                            'controller/statistics',
                                            None,
                                            None,
                                            None)

    fake_params = urllib.parse.parse_qs(
        'user=admin&password=admin&scheme=http&auth=basic')

    def setUp(self):
        super(_Base, self).setUp()
        self.addCleanup(mock.patch.stopall)
        conf = service.prepare_service([], [])
        self.driver = driver.OpenDaylightDriver(conf)
        ks_client = mock.Mock(auth_token='fake_token')
        ks_client.projects.find.return_value = mock.Mock(name='admin',
                                                         id=ADMIN_ID)
        self.ks_client = mock.patch('ceilometer.keystone_client.get_client',
                                    return_value=ks_client).start()
        self.get_statistics = mock.patch(
            'networking_odl.ceilometer.network.statistics.opendaylight_v2.'
            'client.SwitchStatisticsAPIClient.get_statistics',
            return_value=self.switch_data).start()

    def _test_for_meter(self, meter_name, expected_data):
        sample_data = self.driver.get_sample_data(meter_name,
                                                  self.fake_odl_url,
                                                  self.fake_params,
                                                  {})

        self.assertEqual(expected_data, list(sample_data))


class TestOpenDayLightDriverInvalid(_Base):

    switch_data = {"flow_capable_switches": []}

    def test_not_implemented_meter(self):
        sample_data = self.driver.get_sample_data('egg',
                                                  self.fake_odl_url,
                                                  self.fake_params,
                                                  {})
        self.assertIsNone(sample_data)

        sample_data = self.driver.get_sample_data('switch.table.egg',
                                                  self.fake_odl_url,
                                                  self.fake_params,
                                                  {})
        self.assertIsNone(sample_data)

    def test_cache(self):
        cache = {}
        self.driver.get_sample_data('switch',
                                    self.fake_odl_url,
                                    self.fake_params,
                                    cache)
        self.driver.get_sample_data('switch',
                                    self.fake_odl_url,
                                    self.fake_params,
                                    cache)
        self.assertEqual(1, self.get_statistics.call_count)

        cache = {}
        self.driver.get_sample_data('switch',
                                    self.fake_odl_url,
                                    self.fake_params,
                                    cache)
        self.assertEqual(2, self.get_statistics.call_count)

    def test_http_error(self):

        mock.patch(
            'networking_odl.ceilometer.network.statistics.opendaylight_v2.'
            'client.SwitchStatisticsAPIClient.get_statistics',
            side_effect=Exception()).start()

        sample_data = self.driver.get_sample_data('switch',
                                                  self.fake_odl_url,
                                                  self.fake_params,
                                                  {})

        self.assertEqual(0, len(sample_data))

        mock.patch(
            'networking_odl.ceilometer.network.statistics.opendaylight_v2.'
            'client.SwitchStatisticsAPIClient.get_statistics',
            side_effect=[Exception(), self.switch_data]).start()
        cache = {}
        self.driver.get_sample_data('switch',
                                    self.fake_odl_url,
                                    self.fake_params,
                                    cache)

        self.assertIn('network.statistics.opendaylight_v2', cache)


class TestOpenDayLightDriverSimple(_Base):

    switch_data = {
        "flow_capable_switches": [{
            "packet_in_messages_received": "501",
            "packet_out_messages_sent": "300",
            "ports": "1",
            "flow_datapath_id": "55120148545607",
            "switch_port_counters": [{
                "bytes_received": "0",
                "bytes_sent": "0",
                "duration": "600",
                "packets_internal_received": "444",
                "packets_internal_sent": "0",
                "packets_received": "0",
                "packets_received_drop": "0",
                "packets_received_error": "0",
                "packets_sent": "0",
                "port_id": "4",
                "tenant_id": PORT_1_TENANT_ID,
                "uuid": PORT_1_ID
            }],
            "table_counters": [{
                "flow_count": "90",
                "table_id": "0"
            }]
        }]
    }

    def test_meter_switch(self):
        expected_data = [
            (1, "55120148545607",
             {'controller': 'OpenDaylight_V2'},
             ADMIN_ID),
        ]
        self._test_for_meter('switch', expected_data)

    def test_meter_switch_ports(self):
        expected_data = [
            (1, "55120148545607",
             {'controller': 'OpenDaylight_V2'},
             ADMIN_ID)
        ]
        self._test_for_meter('switch.ports', expected_data)

    def test_meter_switch_port(self):
        expected_data = [
            (1, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port', expected_data)

    def test_meter_switch_port_uptime(self):
        expected_data = [
            (600, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.uptime', expected_data)

    def test_meter_switch_port_receive_packets(self):
        expected_data = [
            (0, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.packets', expected_data)

    def test_meter_switch_port_transmit_packets(self):
        expected_data = [
            (0, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.transmit.packets', expected_data)

    def test_meter_switch_port_receive_bytes(self):
        expected_data = [
            (0, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.bytes', expected_data)

    def test_meter_switch_port_transmit_bytes(self):
        expected_data = [
            (0, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.transmit.bytes', expected_data)

    def test_meter_switch_port_receive_drops(self):
        expected_data = [
            (0, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.drops', expected_data)

    def test_meter_switch_port_receive_errors(self):
        expected_data = [
            (0, '55120148545607:4', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 4,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.errors', expected_data)

    def test_meter_port(self):
        expected_data = [
            (1, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port', expected_data)

    def test_meter_port_uptime(self):
        expected_data = [
            (600, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.uptime', expected_data)

    def test_meter_port_receive_packets(self):
        expected_data = [
            (0, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.packets', expected_data)

    def test_meter_port_transmit_packets(self):
        expected_data = [
            (0, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.transmit.packets', expected_data)

    def test_meter_port_receive_bytes(self):
        expected_data = [
            (0, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.bytes', expected_data)

    def test_meter_port_transmit_bytes(self):
        expected_data = [
            (0, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.transmit.bytes', expected_data)

    def test_meter_port_receive_drops(self):
        expected_data = [
            (0, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.drops', expected_data)

    def test_meter_port_receive_errors(self):
        expected_data = [
            (0, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.errors', expected_data)

    def test_meter_switch_table_active_entries(self):
        expected_data = [
            (90, "55120148545607:table:0", {
                'switch': '55120148545607',
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.table.active.entries', expected_data)


class TestOpenDayLightDriverComplex(_Base):

    switch_data = {
        "flow_capable_switches": [{
            "packet_in_messages_received": "501",
            "packet_out_messages_sent": "300",
            "ports": "3",
            "flow_datapath_id": "55120148545607",
            "switch_port_counters": [{
                "bytes_received": "0",
                "bytes_sent": "512",
                "duration": "200",
                "packets_internal_received": "444",
                "packets_internal_sent": "0",
                "packets_received": "10",
                "packets_received_drop": "0",
                "packets_received_error": "0",
                "packets_sent": "0",
                "port_id": "3",
            }, {
                "bytes_received": "9800",
                "bytes_sent": "6540",
                "duration": "150",
                "packets_internal_received": "0",
                "packets_internal_sent": "7650",
                "packets_received": "20",
                "packets_received_drop": "0",
                "packets_received_error": "0",
                "packets_sent": "0",
                "port_id": "2",
                "tenant_id": PORT_2_TENANT_ID,
                "uuid": PORT_2_ID
            }, {
                "bytes_received": "100",
                "bytes_sent": "840",
                "duration": "100",
                "packets_internal_received": "984",
                "packets_internal_sent": "7950",
                "packets_received": "9900",
                "packets_received_drop": "1500",
                "packets_received_error": "1000",
                "packets_sent": "7890",
                "port_id": "1",
                "tenant_id": PORT_1_TENANT_ID,
                "uuid": PORT_1_ID
            }],
            "table_counters": [{
                "flow_count": "90",
                "table_id": "10"
            }, {
                "flow_count": "80",
                "table_id": "20"
            }],
        }, {
            "packet_in_messages_received": "0",
            "packet_out_messages_sent": "0",
            "ports": "0",
            "flow_datapath_id": "55120148545555",
            "table_counters": [{
                "flow_count": "5",
                "table_id": "10"
            }, {
                "flow_count": "3",
                "table_id": "20"
            }],
        }]
    }

    def test_meter_switch(self):
        expected_data = [
            (1, "55120148545607", {
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
            (1, "55120148545555", {
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
        ]

        self._test_for_meter('switch', expected_data)

    def test_meter_switch_ports(self):
        expected_data = [
            (3, "55120148545607", {
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
            (0, "55120148545555", {
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
        ]

        self._test_for_meter('switch.ports', expected_data)

    def test_meter_switch_port(self):
        expected_data = [
            (1, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (1, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (1, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port', expected_data)

    def test_meter_switch_port_uptime(self):
        expected_data = [
            (200, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (150, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (100, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.uptime', expected_data)

    def test_meter_switch_port_receive_packets(self):
        expected_data = [
            (10, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (20, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (9900, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.packets', expected_data)

    def test_meter_switch_port_transmit_packets(self):
        expected_data = [
            (0, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (0, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (7890, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.transmit.packets', expected_data)

    def test_meter_switch_port_receive_bytes(self):
        expected_data = [
            (0, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (9800, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (100, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.bytes', expected_data)

    def test_meter_switch_port_transmit_bytes(self):
        expected_data = [
            (512, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (6540, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (840, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.transmit.bytes', expected_data)

    def test_meter_switch_port_receive_drops(self):
        expected_data = [
            (0, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (0, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (1500, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.drops', expected_data)

    def test_meter_switch_port_receive_errors(self):
        expected_data = [
            (0, "55120148545607:3", {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 3,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (0, '55120148545607:2', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 2,
                'neutron_port_id': PORT_2_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
            (1000, '55120148545607:1', {
                'controller': 'OpenDaylight_V2',
                'port_number_on_switch': 1,
                'neutron_port_id': PORT_1_ID,
                'switch': '55120148545607'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.port.receive.errors', expected_data)

    def test_meter_port(self):
        expected_data = [
            (1, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (1, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port', expected_data)

    def test_meter_port_uptime(self):
        expected_data = [
            (150, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (100, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.uptime', expected_data)

    def test_meter_port_receive_packets(self):
        expected_data = [
            (20, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (9900, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.packets', expected_data)

    def test_meter_port_transmit_packets(self):
        expected_data = [
            (0, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (7890, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.transmit.packets', expected_data)

    def test_meter_port_receive_bytes(self):
        expected_data = [
            (9800, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (100, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.bytes', expected_data)

    def test_meter_port_transmit_bytes(self):
        expected_data = [
            (6540, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (840, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.transmit.bytes', expected_data)

    def test_meter_port_receive_drops(self):
        expected_data = [
            (0, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (1500, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.drops', expected_data)

    def test_meter_port_receive_errors(self):
        expected_data = [
            (0, PORT_2_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_2_TENANT_ID),
            (1000, PORT_1_ID,
             {'controller': 'OpenDaylight_V2'},
             PORT_1_TENANT_ID),
        ]
        self._test_for_meter('port.receive.errors', expected_data)

    def test_meter_switch_table_active_entries(self):
        expected_data = [
            (90, "55120148545607:table:10", {
                'switch': '55120148545607',
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
            (80, "55120148545607:table:20", {
                'switch': '55120148545607',
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
            (5, "55120148545555:table:10", {
                'switch': '55120148545555',
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
            (3, "55120148545555:table:20", {
                'switch': '55120148545555',
                'controller': 'OpenDaylight_V2'
            }, ADMIN_ID),
        ]
        self._test_for_meter('switch.table.active.entries', expected_data)



# Copyright 2016 Intel Corporation.
# Copyright 2016 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from unittest import mock

from pecan import util as p_util

from neutron.tests.unit.testlib_api import SqlTestCaseLight
from neutron_lib import context as neutron_context
from neutron_lib.db import api as db_api
from neutron_lib import fixture as lib_fixtures
from oslo_config import fixture as config_fixture
from oslo_db import exception as db_exc
import sqlalchemy
from sqlalchemy.orm import exc

from networking_odl.common import constants
from networking_odl.db import models


RETRIABLE_EXCEPTIONS = (db_exc.DBDeadlock, exc.StaleDataError,
                        db_exc.DBConnectionError, db_exc.DBDuplicateEntry,
                        db_exc.RetryRequest)

RETRY_INTERVAL = 0.001
RETRY_MAX = 2


class _InnerException(Exception):
    pass


class ODLBaseDbTestCase(SqlTestCaseLight):

    UPDATE_ROW = [constants.ODL_NETWORK, 'id', constants.ODL_UPDATE,
                  {'test': 'data'}]

    def setUp(self):
        super(ODLBaseDbTestCase, self).setUp()
        self.db_context = neutron_context.get_admin_context()
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(completed_rows_retention=-1, group='ml2_odl')
        self._setup_retry_tracker_table()

    def _setup_retry_tracker_table(self):
        metadata = sqlalchemy.MetaData()
        self.retry_table = sqlalchemy.Table(
            'retry_tracker', metadata,
            sqlalchemy.Column(
                'id', sqlalchemy.Integer,
                autoincrement=True,
                primary_key=True,
            ),
        )
        metadata.create_all(self.engine)
        self.addCleanup(metadata.drop_all, self.engine)

        class RetryTracker(object):
            pass

        sqlalchemy.orm.mapper(RetryTracker, self.retry_table)
        self.retry_tracker = RetryTracker

    def _db_cleanup(self):
        self.db_context.session.query(models.OpenDaylightJournal).delete()
        self.db_context.session.query(models.OpenDaylightPeriodicTask).delete()
        row0 = models.OpenDaylightPeriodicTask(
            task='maintenance', state=constants.PENDING)
        row1 = models.OpenDaylightPeriodicTask(
            task='hostconfig', state=constants.PENDING)
        self.db_context.session.merge(row0)
        self.db_context.session.merge(row1)
        self.db_context.session.flush()

    def _test_db_exceptions_handled(self, method, mock_object, expect_retries):
        # NOTE(mpeterson): make retries faster so it doesn't take a lot.
        retry_fixture = lib_fixtures.DBRetryErrorsFixture(
            max_retries=RETRY_MAX, retry_interval=RETRY_INTERVAL)
        retry_fixture.setUp()
        # NOTE(mpeterson): this test is very verbose, disabling logging
        logging.disable(logging.CRITICAL)
        self.addCleanup(logging.disable, logging.NOTSET)
        exceptions = RETRIABLE_EXCEPTIONS

        r_method = getattr(method, '__wrapped__', method)
        r_method_args = p_util.getargspec(r_method).args
        args_number = len(r_method_args) - (2 if r_method_args[0] == 'self'
                                            else 1)
        mock_arg = mock.MagicMock(unsafe=True)
        # NOTE(mpeterson): workarounds for py3 compatibility and behavior
        # expected by particular functions
        mock_arg.__name__ = 'mock_arg'
        mock_arg.retry_count = 1
        mock_arg.__ge__.return_value = True
        mock_arg.__gt__.return_value = True
        mock_arg.__le__.return_value = True
        mock_arg.__lt__.return_value = True
        args = (mock_arg,) * args_number

        def _assertRaises(exceptions, method, context, *args, **kwargs):
            try:
                method(context, *args, **kwargs)
            except Exception as e:
                if not isinstance(e, exceptions):
                    raise e

                # TODO(mpeterson): For now the check with session.is_active is
                # accepted, but when the enginefacade is the only accepted
                # pattern then it should be changed to check that a session is
                # attached to the context
                session = context.session

                is_active = db_api.is_session_active(session)
                if is_active and isinstance(e, _InnerException):
                    self.assertTrue(getattr(e, '_RETRY_EXCEEDED', False))
                return

            exc_names = (tuple(exc.__name__ for exc in exceptions)
                         if hasattr(exceptions, '__iter__') else
                         exceptions.__name__)

            self.fail('%s did not raise %s' % (method.__name__, exc_names))

        try:
            raise _InnerException
        except _InnerException as e:
            _e = e

        expected_retries = RETRY_MAX if expect_retries else 0

        retry_counter = 0
        for exception in exceptions:
            def increase_retry_counter_and_except(*args, **kwargs):
                nonlocal retry_counter
                retry_counter += 1
                self.db_context.session.add(self.retry_tracker())
                self.db_context.session.flush()
                raise exception(_e)
            mock_object.side_effect = increase_retry_counter_and_except

            _assertRaises((exception, _InnerException), method,
                          self.db_context, *args)
            self.assertEqual(expected_retries, mock_object.call_count - 1)
            mock_object.reset_mock()

        retry_fixture.cleanUp()
        return retry_counter

    def _assertRetryCount(self, expected_count):
        actual_count = \
            self.db_context.session.query(self.retry_tracker).count()
        self.assertEqual(expected_count, actual_count)

    def _test_retry_exceptions(self, method, mock_object,
                               assert_transaction=True):
        with mock.patch.object(db_api, 'is_session_active') as sa:
            sa.return_value = False
            retries = self._test_db_exceptions_handled(method, mock_object,
                                                       True)

        if assert_transaction:
            # It should be 0 as long as the retriable method creates save
            # points or transactions, which is the correct behavior
            self._assertRetryCount(0)
            # RETRIABLE * 3 when expect_retries=True since it will retry
            # twice as per the test, plus the original call.
            self.assertEqual(
                len(RETRIABLE_EXCEPTIONS) * (RETRY_MAX + 1),
                retries
            )

        with db_api.CONTEXT_WRITER.using(self.db_context):
            with mock.patch.object(db_api, 'is_session_active') as sa:
                sa.return_value = True
                retries = self._test_db_exceptions_handled(
                    method, mock_object, False
                )
                if assert_transaction:
                    self._assertRetryCount(0)
                    # only once per exception when expect_retries=False
                    self.assertEqual(len(RETRIABLE_EXCEPTIONS), retries)



#
# Copyright (C) 2017 Ericsson India Global Services Pvt Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import webob.exc

from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit import testlib_api

# BGPVPN Table metadata should be imported before
# sqlalchemy metadata.create_all call else tables
# will not be created.
from networking_bgpvpn.neutron.db import bgpvpn_db  # noqa
from networking_bgpvpn.tests.unit.services import test_plugin as bgpvpn_plugin

from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base


class _TestBGPVPNBase(base.OdlTestsBase):
    rds = ['100:1']

    def setUp(self, plugin=None, service_plugins=None,
              ext_mgr=None):
        provider = {
            'service_type': 'BGPVPN',
            'name': 'OpenDaylight',
            'driver': 'networking_odl.bgpvpn.odl_v2.OpenDaylightBgpvpnDriver',
            'default': True
        }
        self.service_providers.return_value = [provider]
        self.plugin_arg = plugin
        self.service_plugin_arg = service_plugins
        self.ext_mgr_arg = ext_mgr
        super(_TestBGPVPNBase, self).setUp()

    def get_ext_managers(self):
        return self.ext_mgr_arg

    def get_plugins(self):
        return self.plugin_arg

    def get_additional_service_plugins(self):
        return self.service_plugin_arg

    def _assert_networks_associated(self, net_ids, bgpvpn):
        response = self.get_odl_resource(odl_const.ODL_BGPVPN, bgpvpn)
        self.assertCountEqual(net_ids,
                              response[odl_const.ODL_BGPVPN]['networks'])

    def _assert_routers_associated(self, router_ids, bgpvpn):
        response = self.get_odl_resource(odl_const.ODL_BGPVPN, bgpvpn)
        self.assertCountEqual(router_ids,
                              response[odl_const.ODL_BGPVPN]['routers'])

    def test_bgpvpn_create(self):
        with self.bgpvpn() as bgpvpn:
            self.assert_resource_created(odl_const.ODL_BGPVPN, bgpvpn)

    def test_bgpvpn_create_with_rds(self):
        with self.bgpvpn(route_distinguishers=self.rds) as bgpvpn:
            response = self.get_odl_resource(odl_const.ODL_BGPVPN, bgpvpn)
            self.assertCountEqual(self.rds,
                                  response[odl_const.ODL_BGPVPN]
                                  ['route_distinguishers'])

    def test_bgpvpn_delete(self):
        with self.bgpvpn(do_delete=False) as bgpvpn:
            self._delete('bgpvpn/bgpvpns', bgpvpn['bgpvpn']['id'])
            self.assertIsNone(
                self.get_odl_resource(odl_const.ODL_BGPVPN, bgpvpn))

    def test_associate_dissociate_net(self):
        with (self.network()) as net1, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn:
            net_id = net1['network']['id']
            bgpvpn_id = bgpvpn['bgpvpn']['id']
            with self.assoc_net(bgpvpn_id, net_id):
                self._assert_networks_associated([net_id], bgpvpn)
            self._assert_networks_associated([], bgpvpn)

    def test_associate_multiple_networks(self):
        with (self.network()) as net1, (self.network()) as net2, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn:
            net_id1 = net1['network']['id']
            net_id2 = net2['network']['id']
            bgpvpn_id = bgpvpn['bgpvpn']['id']
            with self.assoc_net(bgpvpn_id, net_id1), \
                    self.assoc_net(bgpvpn_id, net_id2):
                self._assert_networks_associated([net_id1, net_id2], bgpvpn)

    def test_assoc_multiple_networks_dissoc_one(self):
        with (self.network()) as net1, (self.network()) as net2, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn:
            net_id1 = net1['network']['id']
            net_id2 = net2['network']['id']
            bgpvpn_id = bgpvpn['bgpvpn']['id']
            with self.assoc_net(bgpvpn_id, net_id1):
                with self.assoc_net(bgpvpn_id, net_id2):
                    self._assert_networks_associated([net_id1, net_id2],
                                                     bgpvpn)
                self._assert_networks_associated([net_id1], bgpvpn)

    def test_associate_dissociate_router(self):
        with (self.router(tenant_id=self._tenant_id)) as router, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn:
            router_id = router['router']['id']
            bgpvpn_id = bgpvpn['bgpvpn']['id']
            with self.assoc_router(bgpvpn_id, router_id):
                self._assert_routers_associated([router_id], bgpvpn)
            self._assert_routers_associated([], bgpvpn)

    def test_associate_multiple_routers(self):
        with (self.router(tenant_id=self._tenant_id, name='r1')) as r1, (
                self.router(tenant_id=self._tenant_id, name='r2')) as r2, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn:
            router_id1 = r1['router']['id']
            router_id2 = r2['router']['id']
            bgpvpn_id = bgpvpn['bgpvpn']['id']
            with self.assoc_router(bgpvpn_id, router_id1):
                self._assert_routers_associated([router_id1], bgpvpn)
                with testlib_api.ExpectedException(
                        webob.exc.HTTPClientError) as ctx_manager:
                    with self.assoc_router(bgpvpn_id, router_id2):
                        pass
                self.assertEqual(webob.exc.HTTPBadRequest.code,
                                 ctx_manager.exception.code)
                self._assert_routers_associated([router_id1], bgpvpn)

    def test_assoc_router_multiple_bgpvpns(self):
        with (self.router(tenant_id=self._tenant_id, name='r1')) as router, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn1, (
                self.bgpvpn()) as bgpvpn2:
            router_id = router['router']['id']
            bgpvpn_id_1 = bgpvpn1['bgpvpn']['id']
            bgpvpn_id_2 = bgpvpn2['bgpvpn']['id']
            with (self.assoc_router(bgpvpn_id_1, router_id)), (
                    self.assoc_router(bgpvpn_id_2, router_id)):
                self._assert_routers_associated([router_id], bgpvpn1)
                self._assert_routers_associated([router_id], bgpvpn2)

    def test_associate_router_network(self):
        with (self.router(tenant_id=self._tenant_id)) as router, (
                self.network()) as net1, (
                self.bgpvpn(route_distinguishers=self.rds)) as bgpvpn:
            router_id = router['router']['id']
            net_id = net1['network']['id']
            bgpvpn_id = bgpvpn['bgpvpn']['id']
            with self.assoc_router(bgpvpn_id, router_id), \
                    self.assoc_net(bgpvpn_id, net_id):
                response = self.get_odl_resource(odl_const.ODL_BGPVPN, bgpvpn)
                self.assertCountEqual([router_id],
                                      response[odl_const.ODL_BGPVPN]
                                      ['routers'])
                self.assertCountEqual([net_id],
                                      response[odl_const.ODL_BGPVPN]
                                      ['networks'])


class TestBGPVPNV2Driver(base.V2DriverAdjustment,
                         bgpvpn_plugin.BgpvpnTestCaseMixin,
                         _TestBGPVPNBase, test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']



# Copyright (c) 2013-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from unittest import mock

from networking_odl.common import callback
from networking_odl.common import constants as odl_const
from networking_odl.tests import base

from neutron_lib.callbacks import events
from neutron_lib.callbacks import resources
import testtools

FAKE_ID = 'fakeid'


class ODLCallbackTestCase(testtools.TestCase):

    def setUp(self):
        self.useFixture(base.OpenDaylightRestClientFixture())
        super(ODLCallbackTestCase, self).setUp()
        self._precommit = mock.Mock()
        self._postcommit = mock.Mock()
        self.sgh = callback.OdlSecurityGroupsHandler(self._precommit,
                                                     self._postcommit)

    def _test_callback_precommit_for_sg(self, event, op, sg, sg_id):
        plugin_context_mock = mock.Mock()
        expected_dict = ({resources.SECURITY_GROUP: sg}
                         if sg is not None else None)

        payload = events.DBEventPayload(
            plugin_context_mock, resource_id=sg_id, desired_state=sg)
        self.sgh.sg_callback_precommit(resources.SECURITY_GROUP,
                                       event, None, payload=payload)
        self._precommit.assert_called_with(
            plugin_context_mock, op,
            callback._RESOURCE_MAPPING[resources.SECURITY_GROUP], sg_id,
            expected_dict, payload=payload)

    def _test_callback_postcommit_for_sg(self, event, op, sg, sg_id):
        plugin_context_mock = mock.Mock()
        expected_dict = ({resources.SECURITY_GROUP: sg}
                         if sg is not None else None)
        payload = events.DBEventPayload(
            plugin_context_mock, resource_id=sg_id, desired_state=sg)
        self.sgh.sg_callback_postcommit(resources.SECURITY_GROUP,
                                        event, None, payload=payload)

        self._postcommit.assert_called_with(
            plugin_context_mock, op,
            callback._RESOURCE_MAPPING[resources.SECURITY_GROUP], sg_id,
            expected_dict, payload=payload)

    def test_callback_precommit_sg_create(self):
        sg = mock.Mock()
        sg_id = sg.get('id')
        self._test_callback_precommit_for_sg(
            events.PRECOMMIT_CREATE, odl_const.ODL_CREATE, sg, sg_id)

    def test_callback_postcommit_sg_create(self):
        sg = mock.Mock()
        sg_id = sg.get('id')
        self._test_callback_postcommit_for_sg(
            events.AFTER_CREATE, odl_const.ODL_CREATE, sg, sg_id)

    def test_callback_precommit_sg_update(self):
        self._test_callback_precommit_for_sg(
            events.PRECOMMIT_UPDATE, odl_const.ODL_UPDATE, mock.Mock(),
            FAKE_ID)

    def test_callback_postcommit_sg_update(self):
        self._test_callback_postcommit_for_sg(
            events.AFTER_UPDATE, odl_const.ODL_UPDATE, mock.Mock(), FAKE_ID)

    def test_callback_precommit_sg_delete(self):
        self._test_callback_precommit_for_sg(
            events.PRECOMMIT_DELETE, odl_const.ODL_DELETE, None, FAKE_ID)

    def test_callback_postcommit_sg_delete(self):
        self._test_callback_postcommit_for_sg(
            events.AFTER_DELETE, odl_const.ODL_DELETE, None, FAKE_ID)

    def _test_callback_precommit_for_sg_rules(
            self, event, op, sg_rule, sg_rule_id):
        plugin_context_mock = mock.Mock()
        expected_dict = ({resources.SECURITY_GROUP_RULE: sg_rule}
                         if sg_rule is not None else None)
        payload = events.DBEventPayload(
            plugin_context_mock,
            resource_id=sg_rule_id,
            desired_state=sg_rule
        )
        self.sgh.sg_callback_precommit(resources.SECURITY_GROUP_RULE,
                                       event, None, payload=payload)
        self._precommit.assert_called_with(
            plugin_context_mock, op,
            callback._RESOURCE_MAPPING[resources.SECURITY_GROUP_RULE],
            sg_rule_id, expected_dict, payload=payload)

    def _test_callback_postcommit_for_sg_rules(
            self, event, op, sg_rule, sg_rule_id):
        plugin_context_mock = mock.Mock()
        expected_dict = ({resources.SECURITY_GROUP_RULE: sg_rule}
                         if sg_rule is not None else None)
        payload = events.DBEventPayload(
            plugin_context_mock,
            resource_id=sg_rule_id,
            desired_state=sg_rule
        )
        self.sgh.sg_callback_postcommit(resources.SECURITY_GROUP_RULE,
                                        event, None, payload=payload)

        self._postcommit.assert_called_with(
            plugin_context_mock, op,
            callback._RESOURCE_MAPPING[resources.SECURITY_GROUP_RULE],
            sg_rule_id, expected_dict, payload=payload)

    def test_callback_precommit_sg_rules_create(self):
        rule = mock.Mock()
        rule_id = rule.get('id')
        self._test_callback_precommit_for_sg_rules(
            events.PRECOMMIT_CREATE, odl_const.ODL_CREATE, rule, rule_id)

    def test_callback_postcommit_sg_rules_create(self):
        rule = mock.Mock()
        rule_id = rule.get('id')
        self._test_callback_postcommit_for_sg_rules(
            events.AFTER_CREATE, odl_const.ODL_CREATE, rule, rule_id)

    def test_callback_precommit_sg_rules_delete(self):
        self._test_callback_precommit_for_sg_rules(
            events.PRECOMMIT_DELETE, odl_const.ODL_DELETE, None, FAKE_ID)

    def test_callback_postcommit_sg_rules_delete(self):
        self._test_callback_postcommit_for_sg_rules(
            events.AFTER_DELETE, odl_const.ODL_DELETE, None, FAKE_ID)

    def test_callback_exception(self):

        class TestException(Exception):

            def __init__(self):
                pass

        self._precommit.side_effect = TestException()
        resource = callback._RESOURCE_MAPPING[resources.SECURITY_GROUP_RULE]
        op = callback._OPERATION_MAPPING[events.PRECOMMIT_CREATE]
        rule = mock.Mock()
        rule_id = rule.get('id')
        with mock.patch.object(callback, 'LOG') as log_mock:
            self.assertRaises(TestException,
                              self._test_callback_precommit_for_sg_rules,
                              events.PRECOMMIT_CREATE, odl_const.ODL_CREATE,
                              rule, rule_id)
            log_mock.log.assert_called_with(
                logging.ERROR, callback.LOG_TEMPLATE,
                {'msg': 'Exception from callback', 'op': op,
                 'res_type': resource, 'res_id': rule_id,
                 'res_dict': {odl_const.ODL_SG_RULE: rule},
                 'data': {odl_const.ODL_SG_RULE: rule,
                          'security_group_rule_id': rule_id},
                 'exc_info': True})



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import functools
import os

from neutron.common import utils
from neutron.tests import base
from neutron.tests.common import helpers
from neutron.tests.unit.plugins.ml2 import test_plugin
from oslo_config import cfg
from oslo_config import fixture as config_fixture

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import utils as odl_utils
from networking_odl.db import db
from networking_odl.tests import base as test_base
from networking_odl.tests.unit import test_base_db


class OdlTestsBase(object):
    #  this is stolen from neutron.tests.functional.base
    # This is the directory from which infra fetches log files
    # for functional tests.
    DEFAULT_LOG_DIR = os.path.join(helpers.get_test_log_path(),
                                   'functional-logs')

    def setUp(self):
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(
            url='http://127.0.0.1:8181/controller/nb/v2/neutron',
            group='ml2_odl')
        self.cfg.config(username='admin', group='ml2_odl')
        self.cfg.config(password='admin', group='ml2_odl')
        self.cfg.config(mechanism_drivers=self._mechanism_drivers, group='ml2')
        self.cfg.config(extension_drivers=[
                        'qos', 'port_security'], group='ml2')
        self.client = client.OpenDaylightRestClient.create_client()
        super(OdlTestsBase, self).setUp()
        base.setup_test_logging(
            cfg.CONF, self.DEFAULT_LOG_DIR, "%s.txt" % self.id())

    def setup_parent(self):
        """Perform parent setup with the common plugin configuration class."""
        # Ensure that the parent setup can be called without arguments
        # by the common configuration setUp.
        service_plugins = {'l3_plugin_name': self.l3_plugin}
        service_plugins.update(self.get_additional_service_plugins())
        parent_setup = functools.partial(
            super(test_plugin.Ml2PluginV2TestCase, self).setUp,
            plugin=self.get_plugins(),
            ext_mgr=self.get_ext_managers(),
            service_plugins=service_plugins
        )
        self.useFixture(test_plugin.Ml2ConfFixture(parent_setup))

    def get_plugins(self):
        return test_plugin.PLUGIN_NAME

    def get_ext_managers(self):
        return None

    def get_odl_resource(self, resource_type, resource):
        return self.client.get_resource(
            resource_type, resource[resource_type]['id'])

    def assert_resource_created(self, resource_type, resource):
        odl_resource = self.get_odl_resource(resource_type, resource)
        self.assertIsNotNone(odl_resource)

    def resource_update_test(self, resource_type, resource):
        update_field = 'name'
        update_value = 'bubu'
        resource = self.get_odl_resource(resource_type, resource)
        self.assertNotEqual(update_value,
                            resource[resource_type][update_field])

        self._update(odl_utils.make_url_object(resource_type),
                     resource[resource_type]['id'],
                     {resource_type: {update_field: update_value}})
        resource = self.get_odl_resource(resource_type, resource)
        self.assertEqual(update_value, resource[resource_type][update_field])

    def resource_delete_test(self, resource_type, resource):
        self._delete(odl_utils.make_url_object(resource_type),
                     resource[resource_type]['id'])
        self.assertIsNone(self.get_odl_resource(resource_type, resource))


class V2DriverAdjustment(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        super(V2DriverAdjustment, self).setUp()
        self.useFixture(test_base.JournalWorkerPidFileFixture())

    def get_odl_resource(self, resource_type, resource):
        def no_journal_rows():
            pending_rows = db.get_all_db_rows_by_state(
                self.db_context, odl_const.PENDING)
            processing_rows = db.get_all_db_rows_by_state(
                self.db_context, odl_const.PROCESSING)
            return len(pending_rows) == 0 and len(processing_rows) == 0

        utils.wait_until_true(no_journal_rows, 5, 0.5)

        return super(V2DriverAdjustment, self).get_odl_resource(
            resource_type, resource)



#
# Copyright (C) 2017 Ericsson India Global Services Pvt Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.l2gateway import driver_v2 as driverv2
from networking_odl.tests.unit import base_v2


class OpenDaylightL2GWDriverTestCase(base_v2.OpenDaylightConfigBase):

    def setUp(self):
        super(OpenDaylightL2GWDriverTestCase, self).setUp()
        self.driver = driverv2.OpenDaylightL2gwDriver(service_plugin=None)

    def _get_fake_l2_gateway(self):
        fake_l2_gateway = {
            "tenant_id": "de0a7495-05c4-4be0-b796-1412835c6820",
            "id": "5227c228-6bba-4bbe-bdb8-6942768ff0f1",
            "name": "test-gateway",
            "devices": [
                {
                    "device_name": "switch1",
                    "interfaces": [
                        {
                            "name": "port1",
                            "segmentation_id": [100]
                        },
                        {
                            "name": "port2",
                            "segmentation_id": [151, 152]
                        }
                    ]
                },
                {
                    "device_name": "switch2",
                    "interfaces": [
                        {
                            "name": "port5",
                            "segmentation_id": [200]
                        },
                        {
                            "name": "port6",
                            "segmentation_id": [251, 252]
                        }
                    ]
                }
            ]
        }
        return fake_l2_gateway

    def _get_fake_l2_gateway_connection(self):
        fake_l2_gateway_connection = {
            "tenant_id": "de0a7495-05c4-4be0-b796-1412835c6820",
            "id": "5227c228-6bba-4bbe-bdb8-6942768ff02f",
            "network_id": "be0a7495-05c4-4be0-b796-1412835c6830",
            "default_segmentation_id": 77,
            "l2_gateway_id": "5227c228-6bba-4bbe-bdb8-6942768ff0f1"
        }
        return fake_l2_gateway_connection

    def _assert_op(self, operation, object_type, data, precommit=True):
        row = None
        rows = sorted(db.get_all_db_rows_by_state(self.db_context,
                                                  odl_const.PENDING),
                      key=lambda x: x.seqnum)
        if precommit:
            row = rows[0]
        if precommit:
            self.db_context.session.flush()
            self.assertEqual(operation, row['operation'])
            self.assertEqual(object_type, row['object_type'])
            self.assertEqual(data['id'], row['object_uuid'])
        else:
            self.assertIsNone(row)

    def test_create_l2_gateway(self):
        fake_data = self._get_fake_l2_gateway()
        self.driver.create_l2_gateway_precommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_CREATE, odl_const.ODL_L2GATEWAY,
                        fake_data)
        self.driver.create_l2_gateway_postcommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_CREATE, odl_const.ODL_L2GATEWAY,
                        fake_data, False)

    def test_delete_l2_gateway(self):
        fake_data = self._get_fake_l2_gateway()
        self.driver.delete_l2_gateway_precommit(self.db_context,
                                                fake_data['id'])
        self._assert_op(odl_const.ODL_DELETE, odl_const.ODL_L2GATEWAY,
                        fake_data)
        self.driver.delete_l2_gateway_postcommit(self.db_context,
                                                 fake_data['id'])
        self._assert_op(odl_const.ODL_DELETE, odl_const.ODL_L2GATEWAY,
                        fake_data, False)

    def test_update_l2_gateway(self):
        fake_data = self._get_fake_l2_gateway()
        self.driver.update_l2_gateway_precommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_UPDATE, odl_const.ODL_L2GATEWAY,
                        fake_data)
        self.driver.update_l2_gateway_postcommit(self.db_context, fake_data)
        self._assert_op(odl_const.ODL_UPDATE, odl_const.ODL_L2GATEWAY,
                        fake_data, False)

    def test_create_l2_gateway_connection(self):
        fake_data = self._get_fake_l2_gateway_connection()
        self.driver.create_l2_gateway_connection_precommit(self.db_context,
                                                           fake_data)
        self._assert_op(odl_const.ODL_CREATE,
                        odl_const.ODL_L2GATEWAY_CONNECTION,
                        fake_data)
        self.driver.create_l2_gateway_connection_postcommit(self.db_context,
                                                            fake_data)
        self._assert_op(odl_const.ODL_CREATE,
                        odl_const.ODL_L2GATEWAY_CONNECTION,
                        fake_data, False)

    def test_delete_l2_gateway_connection(self):
        fake_data = self._get_fake_l2_gateway_connection()
        self.driver.delete_l2_gateway_connection_precommit(self.db_context,
                                                           fake_data['id'])
        self._assert_op(odl_const.ODL_DELETE,
                        odl_const.ODL_L2GATEWAY_CONNECTION,
                        fake_data)
        self.driver.delete_l2_gateway_connection_postcommit(self.db_context,
                                                            fake_data['id'])
        self._assert_op(odl_const.ODL_DELETE,
                        odl_const.ODL_L2GATEWAY_CONNECTION,
                        fake_data, False)



# Copyright (C) 2016 Intel Corp. Isaku Yamahata <isaku.yamahata@gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron.tests import base
from neutron_lib import constants as n_const

from networking_odl.common import filters


PROFILE = {"capabilities": ["switchdev"]}
PROFILE_STR = '{"capabilities": ["switchdev"]}'
FAKE_PORT = {'status': 'DOWN',
             'binding:host_id': '',
             'allowed_address_pairs': [],
             'device_owner': 'fake_owner',
             'binding:profile': {"capabilities": ["switchdev"]},
             'fixed_ips': [],
             'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
             'security_groups': [],
             'device_id': 'fake_device',
             'name': '',
             'admin_state_up': True,
             'network_id': 'c13bba05-eb07-45ba-ace2-765706b2d701',
             'tenant_id': 'bad_tenant_id',
             'binding:vif_details': {},
             'binding:vnic_type': 'normal',
             'binding:vif_type': 'unbound',
             'mac_address': '12:34:56:78:21:b6'}


class TestFilters(base.DietTestCase):

    def _check_id(self, resource, project_id):
        filters._populate_project_id_and_tenant_id(resource)
        self.assertIn(resource['project_id'], project_id)
        self.assertIn(resource['tenant_id'], project_id)

    def _test_populate_project_id_and_tenant_id(self, project_id):
        self._check_id({'project_id': project_id}, project_id)
        self._check_id({'tenant_id': project_id}, project_id)
        self._check_id({'project_id': project_id,
                        'tenant_id': project_id}, project_id)

    def test_populate_project_id_and_tenant_id_with_id(self):
        self._test_populate_project_id_and_tenant_id(
            '01234567-890a-bcde-f012-3456789abcde')
        self._test_populate_project_id_and_tenant_id("")

    def test_populate_project_id_and_tenant_id_without_id(self):
        resource = {}
        filters._populate_project_id_and_tenant_id(resource)
        self.assertNotIn('project_id', resource)
        self.assertNotIn('tenant_id', resource)

    def test_populate_project_id_and_tenant_id_with_router(self):
        # test case for OpenDaylightL3RouterPlugin.delete_router()
        # it passes data as dependency_list as list, not dict
        resource0 = ['gw_port_id']
        resource1 = resource0[:]
        filters._populate_project_id_and_tenant_id(resource1)
        self.assertEqual(resource0, resource1)

    def test_populate_project_id_and_tenant_id_with_floatingip(self):
        # test case for OpenDaylightL3RouterPlugin.delete_floatingip()
        # it passes data as dependency_list as list, not dict.
        resource0 = ['router_uuid', 'floatingip_uuid']
        resource1 = resource0[:]
        filters._populate_project_id_and_tenant_id(resource1)
        self.assertEqual(resource0, resource1)

    def test_sgrule_scrub_unknown_protocol_name(self):
        KNOWN_PROTO_NAMES = (n_const.PROTO_NAME_TCP,
                             n_const.PROTO_NAME_UDP,
                             n_const.PROTO_NAME_ICMP,
                             n_const.PROTO_NAME_IPV6_ICMP_LEGACY)
        for protocol_name in KNOWN_PROTO_NAMES:
            self.assertEqual(
                protocol_name,
                filters._sgrule_scrub_unknown_protocol_name(protocol_name))

        self.assertEqual(
            n_const.PROTO_NUM_AH,
            filters._sgrule_scrub_unknown_protocol_name(n_const.PROTO_NAME_AH))
        self.assertEqual("1", filters._sgrule_scrub_unknown_protocol_name("1"))

    def test_sgrule_scrub_icmpv6_name(self):
        for protocol_name in (n_const.PROTO_NAME_ICMP,
                              n_const.PROTO_NAME_IPV6_ICMP,
                              n_const.PROTO_NAME_IPV6_ICMP_LEGACY):
            sgrule = {'ethertype': n_const.IPv6,
                      'protocol': protocol_name}
            filters._sgrule_scrub_icmpv6_name(sgrule)
            self.assertEqual(n_const.PROTO_NAME_IPV6_ICMP_LEGACY,
                             sgrule['protocol'])

    def test_convert_value_to_string(self):
        port = {"binding:profile": PROFILE,
                "other_param": ["some", "values"]}
        filters._convert_value_to_str(port, 'binding:profile')
        self.assertIs(type(port['binding:profile']), str)
        self.assertEqual(port['binding:profile'], PROFILE_STR)
        self.assertIsNot(type(port['other_param']), str)

    def test_convert_value_to_string_unicode(self):
        port = {"binding:profile": {"capabilities": ["switchdev"]}}
        filters._convert_value_to_str(port, "binding:profile")
        self.assertEqual(port["binding:profile"], PROFILE_STR)

    def test_convert_value_to_string_missing_key_is_logged(self):
        port = {}
        with mock.patch.object(filters, 'LOG') as mock_log:
            filters._convert_value_to_str(port, 'invalid_key')
            mock_log.warning.assert_called_once_with(
                "key %s is not present in dict %s", 'invalid_key', port)

    def _filter_port_func_binding_profile_to_string(self, func):
        port = copy.deepcopy(FAKE_PORT)
        func(port)
        self.assertEqual(port["binding:profile"], PROFILE_STR)

    def test_filter_port_create_binding_profile_string(self):
        self._filter_port_func_binding_profile_to_string(
            filters._filter_port_create)

    def test_filter_port_update_binding_profile_string(self):
        self._filter_port_func_binding_profile_to_string(
            filters._filter_port_update)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import testscenarios

from networking_odl.common import constants as odl_const
from networking_odl.dhcp import odl_dhcp_driver
from networking_odl.ml2 import mech_driver_v2
from networking_odl.tests.unit.dhcp import test_odl_dhcp_driver_base

from oslo_config import cfg

load_tests = testscenarios.load_tests_apply_scenarios

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')


class OdlDhcpDriverTestCase(test_odl_dhcp_driver_base.OdlDhcpDriverTestBase):

    def setUp(self):
        super(OdlDhcpDriverTestCase, self).setUp()
        self.cfg.config(enable_dhcp_service=True, group='ml2_odl')
        self.mech = mech_driver_v2.OpenDaylightMechanismDriver()
        self.mech.initialize()

    def test_dhcp_flag_test(self):
        self.assertTrue(cfg.CONF.ml2_odl.enable_dhcp_service)

    def test_dhcp_driver_load(self):
        self.assertTrue(isinstance(self.mech.dhcp_driver,
                                   odl_dhcp_driver.OdlDhcpDriver))

    def test_dhcp_port_create_on_subnet_event(self):

        data = self.get_network_and_subnet_context('10.0.50.0/24', True, True,
                                                   True)
        subnet_context = data['subnet_context']
        mech_driver_v2.OpenDaylightMechanismDriver._record_in_journal(
            subnet_context, odl_const.ODL_SUBNET, odl_const.ODL_CREATE)
        self.mech.journal.sync_pending_entries()

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNotNone(port)

    def test_dhcp_port_create_on_v6subnet_event(self):

        data = self.get_network_and_subnet_context('2001:db8:abcd:0012::0/64',
                                                   True, True, True, False)
        subnet_context = data['subnet_context']
        mech_driver_v2.OpenDaylightMechanismDriver._record_in_journal(
            subnet_context, odl_const.ODL_SUBNET, odl_const.ODL_CREATE)
        self.mech.journal.sync_pending_entries()

        port = self.get_port_id(data['plugin'],
                                data['context'],
                                data['network_id'],
                                data['subnet_id'])
        self.assertIsNone(port)

    def test_dhcp_delete_on_port_update_event(self):

        data = self.get_network_and_subnet_context('10.0.50.0/24', True, True,
                                                   True)
        subnet_context = data['subnet_context']
        plugin = data['plugin']
        self.mech.dhcp_driver.create_or_delete_dhcp_port(subnet_context)
        port_id = self.get_port_id(data['plugin'],
                                   data['context'],
                                   data['network_id'],
                                   data['subnet_id'])

        self.assertIsNotNone(port_id)

        port = plugin.get_port(data['context'], port_id)
        port['fixed_ips'] = []
        ports = {'port': port}
        plugin.update_port(data['context'], port_id, ports)

        mech_driver_v2.OpenDaylightMechanismDriver._record_in_journal(
            subnet_context, odl_const.ODL_PORT, odl_const.ODL_UPDATE, port)
        self.mech.journal.sync_pending_entries()
        port_id = self.get_port_id(data['plugin'],
                                   data['context'],
                                   data['network_id'],
                                   data['subnet_id'])
        self.assertIsNone(port_id)



#!/usr/bin/env python

# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Command line script to set host OVS configurations (it requires ovsctl)

Examples:
    NOTE: bash accepts new line characters between quotes

    To give a full custom json

        python set_ovs_hostconfigs.py --ovs_hostconfigs='{
                "ODL L2": {
                    "allowed_network_types":
                        ["local","vlan", "vxlan","gre"],
                    "bridge_mappings": {"physnet1":"br-ex"}
                    "supported_vnic_types": [
                        {
                            "vnic_type":"normal",
                            "vif_type":"ovs",
                            "vif_details":{}
                         }
                    ],
                },
                "ODL L3": {}
            }'

    To make sure to use system data path (Kernel)

        python set_ovs_hostconfigs.py --noovs_dpdk

    To make sure to use user space data path (vhostuser)

        python set_ovs_hostconfigs.py --ovs_dpdk

    To give bridge mappings

        python --bridge_mapping=physnet1:br-ex,physnet2:br-eth0

"""


import os
import socket
import subprocess  # nosec
import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from networking_odl._i18n import _


LOG = logging.getLogger(__name__)

USERSPACE_DATAPATH_TYPES = ['netdev', 'dpdkvhostuser']

COMMAND_LINE_OPTIONS = [

    cfg.ListOpt(
        'allowed_network_types',
        default=['local', 'flat', 'vlan', 'vxlan', 'gre'],
        help=_("""
            Specifies allowed network types given as a Comma-separated list of
            types.

            Default: --allowed_network_types=local,vlan,vxlan,gre
            """)),

    cfg.DictOpt(
        'bridge_mappings',
        default={},
        help=_("""
            Comma-separated list of <physical_network>:<bridge> tuples mapping
            physical network names to the agent's node-specific Open vSwitch
            bridge names to be used for flat and VLAN networks. The length of
            bridge names should be no more than 11. Each bridge must exist, and
            should have a physical network interface configured as a port. All
            physical networks configured on the server should have mappings to
            appropriate bridges on each agent.

            Note: If you remove a bridge from this mapping, make sure to
            disconnect it from the integration bridge as it won't be managed by
            the agent anymore.

            Default: --bridge_mappings=
            """)),

    cfg.StrOpt(
        'datapath_type',
        choices=['system', 'netdev', 'dpdkvhostuser'],
        default=None,
        help=_("""
            It specifies the OVS data path to use.

            If this value is given then --ovs_dpdk will be ignored.
            If neither this option or --ovs_dpdk are given then it will use a
            valid value for current host.

            Choices: --datapath_type=
                     --datapath_type=system         # kernel data path
                     --datapath_type=netdev         # userspace data path
                     --datapath_type=dpdkvhostuser  # userspace data path

            Default: --datapath_type=netdev         # if support is detected
                     --datapath_type=system         # in all other cases
            """)),

    cfg.StrOpt(
        'host',
        default=socket.gethostname(),  # pylint: disable=no-member
        help=_("""
            It specifies the host name of the target machine.

            Default: --host=$HOSTNAME  # running machine host name
            """)),

    cfg.IPOpt(
        'local_ip',
        help=_("""
            IP address of local overlay (tunnel) network end-point.
            It accepts either an IPv4 or IPv6 address that resides on one
            of the host network interfaces. The IP version of this
            value must match the value of the 'overlay_ip_version'
            option in the ML2 plug-in configuration file on the Neutron
            server node(s).

            Default: local_ip=
            """)),

    cfg.BoolOpt(
        'ovs_dpdk',
        default=None,
        help=_("""
            It uses user-space type of virtual interface (vhostuser) instead of
            the system based one (ovs).

            If this option is not specified it tries to detect vhostuser
            support on running host and in case of positive match it uses it.

            NOTE: if --datapath_type is given then this option is ignored.

            Default:
            """)),

    cfg.BoolOpt(
        'ovs_sriov_offload',
        default=None,
        help=_("""
            It adds SR-IOV virtual interface support to allow ovs hardware
            offload.

            NOTE: This feature should be used with ovs>=2.8.0 and SR-IOV NIC
            which support switchdev mode and tc offload.

            Default:
            """)),

    cfg.StrOpt(
        'ovs_hostconfigs',
        help=_("""
            Fives pre-made host configuration for OpenDaylight as a JSON
            string.

            NOTE: when specified all other options are ignored!

            An entry should look like:
                --ovs_hostconfigs='{
                    "ODL L2": {
                        "allowed_network_types":
                            ["local","vlan", "vxlan","gre"],
                        "bridge_mappings": {"physnet1":"br-ex"}
                        "supported_vnic_types": [
                            {
                                "vnic_type":"normal",
                                "vif_type":"ovs",
                                "vif_details":{}
                             }
                        ],
                    },
                    "ODL L3": {}
                }'

            Default: --ovs_hostconfigs=
            """)),

    cfg.StrOpt(
        'vhostuser_mode',
        choices=['client', 'server'],
        default='client',
        help=_("""
            It specifies the OVS VHostUser mode.

            Choices: --vhostuser_mode=client
                     --vhostuser_mode=server

            Default: --vhostuser_mode=client
            """)),

    cfg.BoolOpt(
        'vhostuser_ovs_plug',
        default=True,
        help=_("""
            Enable VHostUser OVS Plug.

            Default: --vhostuser_ovs_plug
            """)),

    cfg.StrOpt(
        'vhostuser_port_prefix',
        choices=['vhu', 'socket'],
        default='vhu',
        help=_("""
            VHostUser socket port prefix.

            Choices: --vhostuser_socket_dir=vhu
                     --vhostuser_socket_dir=socket

            Default: --vhostuser_socket_dir=vhu
            """)),

    cfg.StrOpt(
        'vhostuser_socket_dir',
        default='/var/run/openvswitch',
        help=_("""
            OVS VHostUser socket directory.

            Default: --vhostuser_socket_dir=/var/run/openvswitch
            """)),
]


def set_ovs_extid_hostconfigs(conf, ovs_vsctl):
    if conf.ovs_hostconfigs:
        json_str = conf.ovs_hostconfigs.replace("\'", "\"")
        LOG.debug("SET-HOSTCONFIGS: JSON String %s", json_str)
        hostconfigs = jsonutils.loads(json_str)

    else:
        uuid = ovs_vsctl.uuid()
        userspace_datapath_types = ovs_vsctl.userspace_datapath_types()
        hostconfigs = _hostconfigs_from_conf(
            conf=conf, uuid=uuid,
            userspace_datapath_types=userspace_datapath_types)

    ovs_vsctl.set_host_name(conf.host)
    for name in sorted(hostconfigs):
        ovs_vsctl.set_host_config(name, hostconfigs[name])

    # for new netvirt
    if conf.local_ip:
        ovs_vsctl.set_local_ip(conf.local_ip)
    if conf.bridge_mappings:
        provider_mappings = ",".join(
            "{}:{}".format(k, v) for k, v in conf.bridge_mappings.items())
        ovs_vsctl.set_provider_mappings(provider_mappings)


def _hostconfigs_from_conf(conf, uuid, userspace_datapath_types):
    vif_type = _vif_type_from_conf(
        conf=conf, userspace_datapath_types=userspace_datapath_types)
    datapath_type = conf.datapath_type or (
        'system' if vif_type == 'ovs' else userspace_datapath_types[0])
    vif_details = _vif_details_from_conf(
        conf=conf, uuid=uuid, vif_type=vif_type)

    host_config = {
        "ODL L2": {
            "allowed_network_types": conf.allowed_network_types,
            "bridge_mappings": conf.bridge_mappings,
            "datapath_type": datapath_type,
            "supported_vnic_types": [
                {
                    "vif_details": vif_details,
                    "vif_type": vif_type,
                    "vnic_type": "normal",
                },
            ]
        }
    }
    if vif_type == 'ovs' and conf.ovs_sriov_offload:
        direct_vnic = {
            "vif_details": vif_details,
            "vif_type": vif_type,
            "vnic_type": "direct",
        }
        host_config["ODL L2"]["supported_vnic_types"].append(direct_vnic)
    return host_config


def _vif_type_from_conf(conf, userspace_datapath_types):

    # take vif_type from datapath_type ------------------------------------
    if conf.datapath_type:
        # take it from  datapath_type
        if conf.datapath_type in USERSPACE_DATAPATH_TYPES:
            if conf.datapath_type not in userspace_datapath_types:
                LOG.warning(
                    "Using user space data path type '%s' even if no "
                    "support was detected.", conf.datapath_type)
            return 'vhostuser'
        else:
            return 'ovs'

    # take vif_type from ovs_dpdk -----------------------------------------
    # pylint: disable=no-else-raise
    if conf.ovs_dpdk is True:
        if userspace_datapath_types:
            return 'vhostuser'

        raise ValueError(_(
            "--ovs_dpdk option was specified but the 'netdev' datapath_type "
            "was not enabled. "
            "To override use option --datapath_type=netdev"))

    elif conf.ovs_dpdk is False:
        return 'ovs'

    # take detected dtype -------------------------------------------------
    if userspace_datapath_types:
        return 'vhostuser'

    return 'ovs'


def _vif_details_from_conf(conf, uuid, vif_type):
    host_addresses = [conf.local_ip or conf.host]
    if vif_type == 'ovs':
        # OVS legacy mode
        return {"uuid": uuid,
                "host_addresses": host_addresses,
                "has_datapath_type_netdev": False,
                "support_vhost_user": False}

    elif vif_type == 'vhostuser':
        # enable VHOSTUSER
        return {"uuid": uuid,
                "host_addresses": host_addresses,
                "has_datapath_type_netdev": True,
                "support_vhost_user": True,
                "port_prefix": conf.vhostuser_port_prefix,
                "vhostuser_socket_dir": conf.vhostuser_socket_dir,
                "vhostuser_ovs_plug": conf.vhostuser_ovs_plug,
                "vhostuser_mode": conf.vhostuser_mode,
                "vhostuser_socket": os.path.join(
                    conf.vhostuser_socket_dir,
                    conf.vhostuser_port_prefix + '$PORT_ID')}
    raise ValueError(_("vif type: '%s' not supported") % vif_type)


def setup_conf(args):
    """setup cmdline options."""
    conf = cfg.ConfigOpts()
    # NOTE, Logging options must be registered before parsing cli
    # options, refer:-
    # https://docs.openstack.org/oslo.log/latest/user/usage.html#oslo-logging-setup-methods
    logging.register_options(conf)

    if '-h' in args or '--help' in args:
        # Prints out script documentation."
        print(__doc__)

    conf.register_cli_opts(COMMAND_LINE_OPTIONS)
    conf(args=args)
    return conf


class OvsVsctl(object):
    """Wrapper class for ovs-vsctl command tool

    """

    COMMAND = 'ovs-vsctl'
    TABLE = 'Open_vSwitch'

    _uuid = None

    def uuid(self):
        uuid = self._uuid
        if uuid is None:
            self._uuid = uuid = self._get('.', '_uuid')
        return uuid

    _datapath_types = None

    def datapath_types(self):
        datapath_types = self._datapath_types
        if datapath_types is None:
            try:
                datapath_types = self._get('.', 'datapath_types')
            except subprocess.CalledProcessError:
                datapath_types = 'system'
            self._datapath_types = datapath_types
        return datapath_types

    _userspace_datapath_types = None

    def userspace_datapath_types(self):
        userspace_datapath_types = self._userspace_datapath_types
        if userspace_datapath_types is None:
            datapath_types = self.datapath_types()
            userspace_datapath_types = tuple(
                datapath_type
                for datapath_type in USERSPACE_DATAPATH_TYPES
                if datapath_type in datapath_types)
            self._userspace_datapath_types = userspace_datapath_types
        return userspace_datapath_types

    def set_host_name(self, host_name):
        self._set_external_ids('odl_os_hostconfig_hostid', host_name)

    def set_host_config(self, name, value):
        self._set_external_ids(
            name='odl_os_hostconfig_config_' + name.lower().replace(' ', '_'),
            value=jsonutils.dumps(value))

    def set_local_ip(self, local_ip):
        self._set_other_config("local_ip", local_ip)

    def set_provider_mappings(self, provider_mappings):
        self._set_other_config("provider_mappings", provider_mappings)

    # --- implementation details ----------------------------------------------

    def _set_external_ids(self, name, value):
        # Refer below for ovs ext-id strings
        # https://review.opendev.org/#/c/309630/
        value = 'external_ids:{}={}'.format(name, value)
        self._set(record=self.uuid(), value=value)

    def _set_other_config(self, name, value):
        value = 'other_config:{}={}'.format(name, value)
        self._set(record=self.uuid(), value=value)

    def _get(self, record, name):
        return self._execute('get', self.TABLE, record, name)

    def _set(self, record, value):
        self._execute('set', self.TABLE, record, value)

    def _execute(self, *args):
        command_line = (self.COMMAND,) + args
        LOG.info(
            "SET-HOSTCONFIGS: Executing cmd: %s", ' '.join(command_line))
        res = subprocess.check_output(command_line).strip()  # nosec
        # Note(lajoskatona): on py3 subprocess.check_output returns back binary
        # to make that consumable we have to decode that.
        if isinstance(res, bytes):
            return res.decode()
        return res


def setup_logging(conf):
    # NOTE, Hacky way to enable logging. oslo log needs other parameters also
    # to register, but neutron.conf is not present on compute node therefore
    # we can not dependant on it. However, nova-compute is present on compute
    # node, so code can depend on oslo log
    # if script does not have write permission in the directory then
    # stack trace appear on console.
    logging.setup(conf, 'networking-odl')
    LOG.info("Logging enabled!")


def main(args=None):
    """Main."""

    if args is None:
        args = sys.argv[1:]

    conf = setup_conf(args)
    setup_logging(conf)

    if os.geteuid() != 0:
        LOG.error('Root permissions are required to configure ovsdb.')
        return 1

    try:
        set_ovs_extid_hostconfigs(conf=conf, ovs_vsctl=OvsVsctl())

    except Exception as ex:  # pylint: disable=broad-except
        LOG.error("Fatal error: %s", ex, exc_info=conf.debug)
        return 1

    else:
        return 0


if __name__ == '__main__':
    sys.exit(main())



#
# Copyright 2017 Ericsson India Global Services Pvt Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urllib

from debtcollector import removals

from oslo_log import log

from ceilometer import keystone_client
from ceilometer.network.statistics import driver
from networking_odl.ceilometer.network.statistics.opendaylight_v2 import client


LOG = log.getLogger(__name__)
INT64_MAX_VALUE = (2 ** 64 / 2 - 1)


@removals.removed_class(
    'OpenDaylightDriver', version='Stein', removal_version='release T',
    message="Ceilometer driver is deprecated and will get removed from n-odl")
class OpenDaylightDriver(driver.Driver):
    """Driver of network info collector from OpenDaylight.

    This driver uses resources in "/etc/ceilometer/polling.yaml".
    Resource requires below conditions:

    * resource is url
    * scheme is "opendaylight.v2"

    This driver can be configured via query parameters.
    Supported parameters:

    * scheme:
      The scheme of request url to OpenDaylight REST API endpoint.
      (default http)
    * auth:
      Auth strategy of http.
      This parameter can be set basic or digest.(default None)
    * user:
      This is username that is used by auth.(default None)
    * password:
      This is password that is used by auth.(default None)

    e.g.::

      opendaylight.v2://127.0.0.1:8080/controller/statistics
      ?auth=basic&user=admin&password=admin&scheme=http

    In this case, the driver send request to below URLs:

      http://127.0.0.1:8080/controller/statistics/flow-capable-switches

    Example JSON response from OpenDaylight
    {
        flow_capable_switches: [{
            "packet_in_messages_received": "501",
            "packet_out_messages_sent": "300",
            "ports": "1",
            "flow_datapath_id": "55120148545607",
            "tenant_id": ADMIN_ID,
            "switch_port_counters": [{
                "bytes_received": "1000",
                "bytes_sent": "1000",
                "duration": "600",
                "packets_internal_received": "100",
                "packets_internal_sent": "200",
                "packets_received": "100",
                "packets_received_drop": "0",
                "packets_received_error": "0",
                "packets_sent": "100",
                "port_id": "4",
                "tenant_id": PORT_1_TENANT_ID,
                "uuid": PORT_1_ID
            }],
            "table_counters": [{
                "flow_count": "90",
                "table_id": "0"
            }]
        }]
    }

    """

    admin_project_id = None

    @staticmethod
    def _get_int_sample(key, statistic, resource_id,
                        resource_meta, tenant_id):
        if key not in statistic:
            return None
        value = int(statistic[key])
        if not (0 <= value <= INT64_MAX_VALUE):
            value = 0
        return value, resource_id, resource_meta, tenant_id

    def _prepare_cache(self, endpoint, params, cache):

        if 'network.statistics.opendaylight_v2' in cache:
            return cache['network.statistics.opendaylight_v2']

        data = {}

        odl_params = {}
        if 'auth' in params:
            odl_params['auth'] = params['auth'][0]
        if 'user' in params:
            odl_params['user'] = params['user'][0]
        if 'password' in params:
            odl_params['password'] = params['password'][0]
        cs = client.Client(self.conf, endpoint, odl_params)

        if not self.admin_project_id:
            try:
                ks_client = keystone_client.get_client(self.conf)
                project = ks_client.projects.find(name='admin')
                if project:
                    self.admin_project_id = project.id
            except Exception:
                LOG.exception('Unable to fetch admin tenant id')
                cache['network.statistics.opendaylight_v2'] = data
                return data

        try:
            # get switch statistics
            data['switch'] = cs.switch_statistics.get_statistics()
            data['admin_tenant_id'] = self.admin_project_id
        except client.OpenDaylightRESTAPIFailed:
            LOG.exception('OpenDaylight REST API Failed. ')
        except Exception:
            LOG.exception('Failed to connect to OpenDaylight'
                          ' REST API')

        cache['network.statistics.opendaylight_v2'] = data

        return data

    def get_sample_data(self, meter_name, parse_url, params, cache):

        extractor = self._get_extractor(meter_name)
        if extractor is None:
            # The way to getting meter is not implemented in this driver or
            # OpenDaylight REST API has not api to getting meter.
            return None

        iter = self._get_iter(meter_name)
        if iter is None:
            # The way to getting meter is not implemented in this driver or
            # OpenDaylight REST API has not api to getting meter.
            return None

        parts = urllib.parse.ParseResult(params.get('scheme', ['http'])[0],
                                         parse_url.netloc,
                                         parse_url.path,
                                         None,
                                         None,
                                         None)
        endpoint = urllib.parse.urlunparse(parts)

        data = self._prepare_cache(endpoint, params, cache)

        samples = []
        if data:
            for sample in iter(extractor, data):
                if sample is not None:
                    # set controller name to resource_metadata
                    sample[2]['controller'] = 'OpenDaylight_V2'
                    samples.append(sample)

        return samples

    def _get_iter(self, meter_name):
        if meter_name == 'switch' or meter_name == 'switch.ports':
            return self._iter_switch
        elif meter_name.startswith('switch.table'):
            return self._iter_table
        elif meter_name.startswith('switch.port'):
            return self._iter_switch_port
        elif meter_name.startswith('port'):
            return self._iter_port
        return None

    def _get_extractor(self, meter_name):
        if (meter_name == 'switch.port' or
                meter_name.startswith('switch.port.')):
            meter_name = meter_name.split('.', 1)[1]
        method_name = '_' + meter_name.replace('.', '_')
        return getattr(self, method_name, None)

    @staticmethod
    def _iter_switch(extractor, data):
        for switch in data['switch']['flow_capable_switches']:
            yield (extractor(switch, switch['flow_datapath_id'], {},
                             (switch.get('tenant_id') or
                              data['admin_tenant_id'])))

    @staticmethod
    def _switch(statistic, resource_id,
                resource_meta, tenant_id):
        return 1, resource_id, resource_meta, tenant_id

    @staticmethod
    def _switch_ports(statistic, resource_id,
                      resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'ports', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _iter_switch_port(extractor, data):
        for switch in data['switch']['flow_capable_switches']:
            if 'switch_port_counters' in switch:
                switch_id = switch['flow_datapath_id']
                tenant_id = (switch.get('tenant_id') or
                             data['admin_tenant_id'])
                for port_statistic in switch['switch_port_counters']:
                    port_id = int(port_statistic['port_id'])
                    resource_id = '%s:%d' % (switch_id, port_id)
                    resource_meta = {'switch': switch_id,
                                     'port_number_on_switch': port_id}
                    if 'uuid' in port_statistic:
                        neutron_port_id = port_statistic['uuid']
                        resource_meta['neutron_port_id'] = neutron_port_id
                    yield extractor(port_statistic, resource_id,
                                    resource_meta, tenant_id)

    @staticmethod
    def _iter_port(extractor, data):
        resource_meta = {}
        for switch in data['switch']['flow_capable_switches']:
            if 'switch_port_counters' in switch:
                for port_statistic in switch['switch_port_counters']:
                    if 'uuid' in port_statistic:
                        resource_id = port_statistic['uuid']
                        tenant_id = port_statistic.get('tenant_id')
                        yield extractor(
                            port_statistic, resource_id, resource_meta,
                            tenant_id or data['admin_tenant_id'])

    @staticmethod
    def _port(statistic, resource_id, resource_meta, tenant_id):
        return 1, resource_id, resource_meta, tenant_id

    @staticmethod
    def _port_uptime(statistic, resource_id,
                     resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'duration', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _port_receive_packets(statistic, resource_id,
                              resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'packets_received', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _port_transmit_packets(statistic, resource_id,
                               resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'packets_sent', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _port_receive_bytes(statistic, resource_id,
                            resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'bytes_received', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _port_transmit_bytes(statistic, resource_id,
                             resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'bytes_sent', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _port_receive_drops(statistic, resource_id,
                            resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'packets_received_drop', statistic, resource_id,
            resource_meta, tenant_id)

    @staticmethod
    def _port_receive_errors(statistic, resource_id,
                             resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'packets_received_error', statistic,
            resource_id, resource_meta, tenant_id)

    @staticmethod
    def _iter_table(extractor, data):
        for switch_statistic in data['switch']['flow_capable_switches']:
            if 'table_counters' in switch_statistic:
                switch_id = switch_statistic['flow_datapath_id']
                tenant_id = (switch_statistic.get('tenant_id') or
                             data['admin_tenant_id'])
                for table_statistic in switch_statistic['table_counters']:
                    resource_meta = {'switch': switch_id}
                    resource_id = ("%s:table:%d" %
                                   (switch_id,
                                    int(table_statistic['table_id'])))
                    yield extractor(table_statistic, resource_id,
                                    resource_meta, tenant_id)

    @staticmethod
    def _switch_table_active_entries(statistic, resource_id,
                                     resource_meta, tenant_id):
        return OpenDaylightDriver._get_int_sample(
            'flow_count', statistic, resource_id,
            resource_meta, tenant_id)



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const
from neutron_lib.plugins.ml2 import api
from oslo_log import log

from networking_odl.ml2 import port_binding


LOG = log.getLogger(__name__)


class LegacyPortBindingManager(port_binding.PortBindingController):

    def __init__(self):
        self.supported_vnic_types = [portbindings.VNIC_NORMAL]
        self.vif_details = {portbindings.CAP_PORT_FILTER: True}

    def bind_port(self, port_context):
        """Set binding for all valid segments

        """
        vnic_type = port_context.current.get(portbindings.VNIC_TYPE,
                                             portbindings.VNIC_NORMAL)
        if vnic_type not in self.supported_vnic_types:
            LOG.debug("Refusing to bind due to unsupported vnic_type: %s",
                      vnic_type)
            return

        valid_segment = None
        for segment in port_context.segments_to_bind:
            if self._check_segment(segment):
                valid_segment = segment
                break

        if valid_segment:
            vif_type = self._get_vif_type(port_context)
            LOG.debug("Bind port %(port)s on network %(network)s with valid "
                      "segment %(segment)s and VIF type %(vif_type)r.",
                      {'port': port_context.current['id'],
                       'network': port_context.network.current['id'],
                       'segment': valid_segment, 'vif_type': vif_type})

            port_context.set_binding(
                valid_segment[api.ID], vif_type,
                self.vif_details,
                status=n_const.PORT_STATUS_ACTIVE)

    def _check_segment(self, segment):
        """Verify a segment is valid for the OpenDaylight MechanismDriver.

        Verify the requested segment is supported by ODL and return True or
        False to indicate this to callers.
        """

        network_type = segment[api.NETWORK_TYPE]
        return network_type in [n_const.TYPE_FLAT, n_const.TYPE_LOCAL,
                                n_const.TYPE_GRE, n_const.TYPE_VXLAN,
                                n_const.TYPE_VLAN]

    def _get_vif_type(self, port_context):
        """Get VIF type string for given PortContext

        Dummy implementation: it always returns following constant.
        neutron_lib.api.definitions.portbindings.VIF_TYPE_OVS
        """

        return portbindings.VIF_TYPE_OVS



# Copyright (c) 2014 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from networking_odl._i18n import _


odl_opts = [
    cfg.StrOpt('url',
               help=_("HTTP URL of OpenDaylight REST interface.")),
    cfg.StrOpt('username',
               help=_("HTTP username for authentication.")),
    cfg.StrOpt('password', secret=True,
               help=_("HTTP password for authentication.")),
    cfg.IntOpt('timeout', default=10,
               help=_("HTTP timeout in seconds.")),
    cfg.IntOpt('session_timeout', default=30,
               help=_("Tomcat session timeout in minutes.")),
    cfg.FloatOpt('sync_timeout', default=10,
                 help=_("Sync thread timeout in seconds or fraction.")),
    cfg.IntOpt('retry_count', default=5,
               help=_("Number of times to retry a row before failing.")),
    cfg.IntOpt('maintenance_interval', default=300,
               help=_("Journal maintenance operations interval in seconds.")),
    cfg.IntOpt('completed_rows_retention', default=0,
               help=_("Time to keep completed rows (in seconds)."
                      "For performance reasons it's not recommended to "
                      "change this from the default value (0) which "
                      "indicates completed rows aren't kept."
                      "This value will be checked every maintenance_interval "
                      "by the cleanup thread. To keep completed rows "
                      "indefinitely, set the value to -1")),
    cfg.BoolOpt('enable_lightweight_testing',
                default=False,
                help=_('Test without real ODL.')),
    cfg.StrOpt('port_binding_controller',
               default='pseudo-agentdb-binding',
               help=_('Name of the controller to be used for port binding.')),
    cfg.IntOpt('processing_timeout', default='100',
               help=_("Time in seconds to wait before a "
                      "processing row is marked back to pending.")),
    cfg.StrOpt('odl_hostconf_uri',
               help=_("Path for ODL host configuration REST interface"),
               default="/restconf/operational/neutron:neutron/hostconfigs"),
    cfg.IntOpt('restconf_poll_interval', default=30,
               help=_("Poll interval in seconds for getting ODL hostconfig")),
    cfg.BoolOpt('enable_websocket_pseudo_agentdb', default=False,
                help=_('Enable websocket for pseudo-agent-port-binding.')),
    cfg.IntOpt('odl_features_retry_interval', default=5,
               help=_("Wait this many seconds before retrying the odl features"
                      " fetch")),
    cfg.ListOpt('odl_features',
                help='A list of features supported by ODL.'),
    cfg.StrOpt('odl_features_json',
               help='Features supported by ODL, in the json format returned'
               'by ODL. Note: This config option takes precedence over'
               'odl_features.'),
    cfg.BoolOpt('enable_dhcp_service', default=False,
                help=_('Enables the networking-odl driver to supply special'
                       ' neutron ports of "dhcp" type to OpenDaylight'
                       ' Controller for its use in providing DHCP Service.')),
]

cfg.CONF.register_opts(odl_opts, "ml2_odl")


def list_opts():
    return [('ml2_odl', odl_opts)]



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import threading
from unittest import mock

from neutron.common import utils
from neutron_lib import context

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.db import models
from networking_odl.journal import periodic_task
from networking_odl.tests.unit import test_base_db


TEST_TASK_NAME = 'test-maintenance'
TEST_TASK_INTERVAL = 0.1


class PeriodicTaskThreadTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        super(PeriodicTaskThreadTestCase, self).setUp()
        row = models.OpenDaylightPeriodicTask(task=TEST_TASK_NAME,
                                              state=odl_const.PENDING)
        self.db_context.session.add(row)
        self.db_context.session.flush()

        self.thread = periodic_task.PeriodicTask(TEST_TASK_NAME,
                                                 TEST_TASK_INTERVAL)
        self.addCleanup(self.thread.cleanup)

    def test__execute_op_no_exception(self):
        with mock.patch.object(periodic_task, 'LOG') as mock_log:
            operation = mock.MagicMock()
            operation.__name__ = "test"
            self.thread.register_operation(operation)
            self.thread._execute_op(operation, self.db_context)
            operation.assert_called()
            mock_log.info.assert_called()
            mock_log.exception.assert_not_called()

    def test__execute_op_with_exception(self):
        with mock.patch.object(periodic_task, 'LOG') as mock_log:
            operation = mock.MagicMock(side_effect=Exception())
            operation.__name__ = "test"
            self.thread._execute_op(operation, self.db_context)
            mock_log.exception.assert_called()

    def test_thread_works(self):
        callback_event = threading.Event()
        count = 0

        def callback_op(*args):
            nonlocal count
            count += 1

            # The following should be true on the second call, so we're making
            # sure that the thread runs more than once.
            if count > 1:
                callback_event.set()

        self.thread.register_operation(callback_op)
        self.thread.start()

        # Make sure the callback event was called and not timed out
        self.assertTrue(callback_event.wait(timeout=5))

    def test_thread_continues_after_exception(self):
        exception_event = threading.Event()
        callback_event = threading.Event()

        def exception_op(*args):
            if not exception_event.is_set():
                exception_event.set()
                raise Exception()

        def callback_op(*args):
            callback_event.set()

        for op in [exception_op, callback_op]:
            self.thread.register_operation(op)

        self.thread.start()

        # Make sure the callback event was called and not timed out
        self.assertTrue(callback_event.wait(timeout=5))

    def test_multiple_thread_work(self):
        self.thread1 = periodic_task.PeriodicTask(TEST_TASK_NAME + '1',
                                                  TEST_TASK_INTERVAL)
        callback_event = threading.Event()
        callback_event1 = threading.Event()
        self.addCleanup(self.thread1.cleanup)

        def callback_op(*args):
            callback_event.set()

        def callback_op1(*args):
            callback_event1.set()

        self.thread.register_operation(callback_op)
        self.thread.register_operation(callback_op1)
        self.thread.start()
        self.assertTrue(callback_event.wait(timeout=5))

        self.thread1.start()
        self.assertTrue(callback_event1.wait(timeout=5))

    @mock.patch.object(db, "was_periodic_task_executed_recently")
    def test_back_to_back_job(self, mock_status_method):
        callback_event = threading.Event()
        continue_event = threading.Event()

        def callback_op(*args):
            callback_event.set()

        return_value = True

        def continue_(*args, **kwargs):
            continue_event.set()
            return return_value
        mock_status_method.side_effect = continue_

        self.thread.register_operation(callback_op)
        msg = ("Periodic %s task executed after periodic "
               "interval Skipping execution.")
        with mock.patch.object(periodic_task.LOG, 'info') as mock_log_info:
            self.thread.start()
            self.assertTrue(continue_event.wait(timeout=1))
            continue_event.clear()
            mock_log_info.assert_called_with(msg, TEST_TASK_NAME)
            self.assertFalse(callback_event.is_set())
            self.assertTrue(continue_event.wait(timeout=1))
            continue_event.clear()
            mock_log_info.assert_called_with(msg, TEST_TASK_NAME)
            return_value = False
            self.assertTrue(callback_event.wait(timeout=2))

    def test_set_operation_retries_exceptions(self):
        with mock.patch.object(db, 'update_periodic_task') as m:
            self._test_retry_exceptions(self.thread._set_operation, m)

    def test_lock_task_retries_exceptions(self):
        with mock.patch.object(db, 'lock_periodic_task') as m:
            self._test_retry_exceptions(self.thread._lock_task, m)

    def test_clear_and_unlock_task_retries_exceptions(self):
        with mock.patch.object(db, 'update_periodic_task') as m:
            self._test_retry_exceptions(self.thread._clear_and_unlock_task, m)

    @mock.patch.object(db, "was_periodic_task_executed_recently",
                       return_value=False)
    def test_no_multiple_executions_simultaneously(self, mock_exec_recently):
        continue_event = threading.Event()
        trigger_event = threading.Event()
        count = 0

        def wait_until_event(context):
            nonlocal count
            trigger_event.set()
            if continue_event.wait(2):
                count += 1

        self.thread.register_operation(wait_until_event)

        def task_locked():
            session = self.db_context.session
            row = (session.query(models.OpenDaylightPeriodicTask)
                          .filter_by(state=odl_const.PROCESSING,
                                     task=TEST_TASK_NAME)
                          .one_or_none())
            return (row is not None)

        self.thread.start()
        utils.wait_until_true(trigger_event.is_set, 5, 0.01)
        self.assertEqual(count, 0)
        self.assertTrue(task_locked())

        self.thread.execute_ops()
        self.assertEqual(count, 0)
        self.assertTrue(task_locked())

        continue_event.set()
        trigger_event.clear()
        utils.wait_until_true(trigger_event.is_set, 5, 0.01)
        self.thread.cleanup()
        self.assertFalse(task_locked())
        self.assertGreaterEqual(count, 1)

    @mock.patch.object(db, "was_periodic_task_executed_recently",
                       return_value=True)
    def test_forced_execution(self, mock_status_method):
        operation = mock.MagicMock()
        operation.__name__ = "test"
        self.thread.register_operation(operation)

        self.thread.execute_ops(forced=True)

        operation.assert_called()

    @mock.patch.object(db, "was_periodic_task_executed_recently",
                       return_value=True)
    def test_context_is_passed_as_args(self, _):
        operation = mock.MagicMock()
        operation.__name__ = 'test'
        self.thread.register_operation(operation)

        self.thread.execute_ops(forced=True)

        # This tests that only ONE args is passed, and no kwargs
        operation.assert_called_with(mock.ANY)

        # This tests that it's a context
        kall = operation.call_args
        args, kwargs = kall
        self.assertIsInstance(args[0], context.Context)



# Copyright (c) 2017 NEC Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from neutron_lib.plugins import directory

from networking_odl.common import constants
from networking_odl.common import exceptions
from networking_odl.db import db
from networking_odl.journal import base_driver
from networking_odl.tests.unit.journal import helper
from networking_odl.tests.unit import test_base_db


class BaseDriverTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        super(BaseDriverTestCase, self).setUp()
        self.test_driver = helper.TestDriver()
        self.plugin = helper.TestPlugin()
        directory.add_plugin(helper.TEST_PLUGIN, self.plugin)
        self.addCleanup(directory.add_plugin, helper.TEST_PLUGIN, None)

    def test_get_resource_driver(self):
        for resource, resource_suffix in self.test_driver.RESOURCES.items():
            driver = base_driver.get_driver(resource)
            self.assertEqual(driver, self.test_driver)
            self.assertEqual(driver.plugin_type, helper.TEST_PLUGIN)
            self.assertEqual(self.test_driver.RESOURCES.get(resource),
                             resource_suffix)

    def non_existing_plugin_cleanup(self):
        self.test_driver.plugin_type = helper.TEST_PLUGIN

    def test_non_existing_plugin(self):
        self.test_driver.plugin_type = helper.INVALID_PLUGIN
        self.addCleanup(self.non_existing_plugin_cleanup)
        self.assertIsNone(self.test_driver.plugin)

    def test_get_non_existing_resource_driver(self):
        self.assertRaises(exceptions.ResourceNotRegistered,
                          base_driver.get_driver, helper.INVALID_RESOURCE)

    def test_get_resources_for_full_sync(self):
        received_resources = self.test_driver.get_resources_for_full_sync(
            self.db_context,
            helper.TEST_RESOURCE1)
        resources = self.plugin.get_test_resource1s(self.db_context)
        for resource in resources:
            self.assertIn(resource, received_resources)

    def test_get_non_existing_resources_for_full_sync(self):
        self.assertRaises(exceptions.UnsupportedResourceType,
                          self.test_driver.get_resources_for_full_sync,
                          self.db_context, helper.INVALID_RESOURCE)

    def test_get_resource(self):
        row = db.create_pending_row(self.db_context, helper.TEST_RESOURCE1,
                                    helper.TEST_UUID, constants.ODL_CREATE,
                                    {'id': helper.TEST_UUID})
        resource = self.test_driver.get_resource_for_recovery(self.db_context,
                                                              row)

        self.assertEqual(resource['id'], helper.TEST_UUID)

    def test_get_unsupported_resource(self):
        row = db.create_pending_row(self.db_context, helper.INVALID_RESOURCE,
                                    helper.TEST_UUID, constants.ODL_CREATE,
                                    {'id': helper.TEST_UUID})

        self.assertRaises(exceptions.PluginMethodNotFound,
                          self.test_driver.get_resource_for_recovery,
                          self.db_context, row)



# Copyright (c) 2017 Brocade Communication Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest.mock import patch

from neutron_lib.db import api as db_api

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.sfc.flowclassifier import sfc_flowclassifier_v2 as sfc_fc
from networking_odl.tests import base as odl_base
from networking_odl.tests.unit import base_v2
from networking_odl.tests.unit.sfc import constants as sfc_const


class TestOpenDaylightSFCFlowClassifierDriverV2(
        base_v2.OpenDaylightConfigBase):

    def setUp(self):
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        super(TestOpenDaylightSFCFlowClassifierDriverV2, self).setUp()
        self.handler = sfc_fc.OpenDaylightSFCFlowClassifierDriverV2()
        self.handler.initialize()

    def _get_mock_context(self):
        mocked_fc_context = patch(
            'networking_sfc.services.flowclassifier.common.context'
            '.FlowClassifierContext').start().return_value

        mocked_fc_context.current = sfc_const.FAKE_FLOW_CLASSIFIER
        mocked_fc_context.session = self.db_context.session
        mocked_fc_context._plugin_context = mocked_fc_context
        return mocked_fc_context

    def _call_operation_object(self, operation, timing):
        method = getattr(self.handler,
                         '%s_flow_classifier_%s' % (operation, timing))
        method(self._get_mock_context())

    def _test_event(self, operation, timing):
        with db_api.CONTEXT_WRITER.using(self.db_context):
            self._call_operation_object(operation, timing)
            if timing == 'precommit':
                self.db_context.session.flush()
            row = db.get_oldest_pending_db_row_with_lock(self.db_context)

            if timing == 'precommit':
                self.assertEqual(operation, row['operation'])
                self.assertEqual(
                    odl_const.ODL_SFC_FLOW_CLASSIFIER, row['object_type'])
            elif timing == 'after':
                self.assertIsNone(row)

    # TODO(yamahata): utilize test scenarios
    def test_create_flow_classifier_precommit(self):
        self._test_event("create", "precommit")

    def test_create_flow_classifier_postcommit(self):
        self._test_event("create", "postcommit")

    def test_update_flow_classifier_precommit(self):
        self._test_event("update", "precommit")

    def test_update_flow_classifier_postcommit(self):
        self._test_event("update", "postcommit")

    def test_delete_flow_classifier_precommit(self):
        self._test_event("delete", "precommit")

    def test_delete_flow_classifier_postcommit(self):
        self._test_event("delete", "postcommit")



# Copyright (c) 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib.db import constants as db_const
from neutron_lib.plugins import constants as nlib_const
from neutron_lib.services.qos import base
from neutron_lib.services.qos import constants as qos_consts
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_odl.common import constants as odl_const
from networking_odl.common import odl_features
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.qos import qos_utils

LOG = logging.getLogger(__name__)

VIF_TYPES = [portbindings.VIF_TYPE_OVS, portbindings.VIF_TYPE_VHOST_USER]
VNIC_TYPES = [portbindings.VNIC_NORMAL]

QOS_RESOURCES = {
    odl_const.ODL_QOS_POLICY: odl_const.ODL_QOS_POLICIES
}

DEFAULT_QOS_RULES = {
    qos_consts.RULE_TYPE_BANDWIDTH_LIMIT: {
        qos_consts.MAX_KBPS: {
            'type:range': [0, db_const.DB_INTEGER_MAX_VALUE]},
        qos_consts.MAX_BURST: {
            'type:range': [0, db_const.DB_INTEGER_MAX_VALUE]},
        qos_consts.DIRECTION: {
            'type:values': [constants.EGRESS_DIRECTION]}
    }
}

QOS_RULES = 'qos-rules'


class OpenDaylightQosDriver(base.DriverBase):

    """OpenDaylight QOS Driver

    This code is backend implementation for OpenDaylight Qos
    driver for Openstack Neutron.
    """

    @staticmethod
    def create():
        try:
            supported_rules = odl_features.get_config(QOS_RULES)
        except KeyError:
            supported_rules = DEFAULT_QOS_RULES

        return OpenDaylightQosDriver(supported_rules=supported_rules)

    def __init__(self, supported_rules,
                 name='OpenDaylightQosDriver',
                 vif_types=VIF_TYPES,
                 vnic_types=VNIC_TYPES,
                 requires_rpc_notifications=False):
        super(OpenDaylightQosDriver, self).__init__(
            name, vif_types, vnic_types, supported_rules,
            requires_rpc_notifications)
        LOG.debug("Initializing OpenDaylight Qos driver")
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(nlib_const.QOS, QOS_RESOURCES)

    def _record_in_journal(self, context, op_const, qos_policy):
        data = qos_utils.convert_rules_format(qos_policy.to_dict())
        journal.record(context, odl_const.ODL_QOS_POLICY,
                       data['id'], op_const, data)

    @log_helpers.log_method_call
    def create_policy_precommit(self, context, policy):
        self._record_in_journal(context, odl_const.ODL_CREATE, policy)

    @log_helpers.log_method_call
    def update_policy_precommit(self, context, policy):
        self._record_in_journal(context, odl_const.ODL_UPDATE, policy)

    @log_helpers.log_method_call
    def delete_policy_precommit(self, context, policy):
        self._record_in_journal(context, odl_const.ODL_DELETE, policy)

    @log_helpers.log_method_call
    def create_policy(self, context, policy):
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    def update_policy(self, context, policy):
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    def delete_policy(self, context, policy):
        self.journal.set_sync_event()



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.services.trunk import models
from neutron_lib.callbacks import events
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory
from neutron_lib.services.trunk import constants as trunk_consts

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.tests.unit import base_v2
from networking_odl.trunk import trunk_driver_v2 as trunk_driver


FAKE_TRUNK = {
    'status': 'ACTIVE',
    'sub_ports': [{'segmentation_type': 'vlan',
                   'port_id': 'fake_port_id',
                   'segmentation_id': 101},
                  {'segmentation_type': 'vlan',
                   'port_id': 'fake_port_id',
                   'segmentation_id': 102}],
    'name': 'trunk0',
    'admin_state_up': 'true',
    'tenant_id': 'fake_tenant_id',
    'updated_at': '2016-11-16T10:17:44Z',
    'revision_number': 2,
    'project_id': 'fake_project_id',
    'port_id': 'fake_port_id',
    'id': 'fake_id',
    'description': 'fake trunk port'}

FAKE_PARENT = {
    'id': 'fake_parent_id',
    'tenant_id': 'fake_tenant_id',
    'name': 'parent_port',
    'admin_state_up': 'true',
    'status': 'ACTIVE'}


class TestTrunkHandler(base_v2.OpenDaylightConfigBase):
    def setUp(self):
        super(TestTrunkHandler, self).setUp()
        self.handler = (trunk_driver.
                        OpenDaylightTrunkHandlerV2())

    def _fake_trunk_payload(self):
        payload = events.DBEventPayload(
            self.db_context, resource_id='fake_id',
            states=(mock.Mock(return_value=FAKE_TRUNK),
                    mock.Mock(return_value=FAKE_TRUNK),))
        payload.latest_state.status = trunk_consts.TRUNK_DOWN_STATUS
        payload.latest_state.to_dict = mock.Mock(return_value=FAKE_TRUNK)
        payload.states[0].status = trunk_consts.TRUNK_DOWN_STATUS
        payload.states[0].to_dict = mock.Mock(return_value=FAKE_TRUNK)
        return payload

    def _call_operation_object(self, operation, timing, fake_payload):
        method = getattr(self.handler, 'trunk_%s_%s' % (operation, timing))
        method(mock.ANY, mock.ANY, mock.ANY, fake_payload)

    def _test_event(self, operation, timing):
        with db_api.CONTEXT_WRITER.using(self.db_context):
            fake_payload = self._fake_trunk_payload()
            self._call_operation_object(operation, timing, fake_payload)
            if timing == 'precommit':
                self.db_context.session.flush()

            row = db.get_oldest_pending_db_row_with_lock(self.db_context)

            if timing == 'precommit':
                self.assertEqual(operation, row['operation'])
                self.assertEqual(odl_const.ODL_TRUNK, row['object_type'])
                self.assertEqual(fake_payload.resource_id, row['object_uuid'])
            elif timing == 'after':
                self.assertIsNone(row)

    def test_trunk_create_precommit(self):
        self._test_event("create", "precommit")

    def test_trunk_create_postcommit(self):
        self._test_event("create", "postcommit")

    def test_trunk_update_precommit(self):
        self._test_event("update", "precommit")

    def test_trunk_update_postcommit(self):
        self._test_event("update", "postcommit")

    def test_trunk_delete_precommit(self):
        self._test_event("delete", "precommit")

    def test_trunk_delete_postcommit(self):
        self._test_event("delete", "postcommit")

    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_set_subport_status')
    def test_trunk_subports_set_status_create_parent_active(
            self, mock_set_subport_status):
        resource = resources.SUBPORTS
        event_type = events.AFTER_CREATE
        fake_payload = self._fake_trunk_payload()
        core_plugin = directory.get_plugin()

        fake_payload.subports = [models.SubPort(port_id='fake_port_id',
                                                segmentation_id=101,
                                                segmentation_type='vlan',
                                                trunk_id='fake_id')]
        parent_port = FAKE_PARENT

        with mock.patch.object(core_plugin, '_get_port') as gp:
            gp.return_value = parent_port
            self.handler.trunk_subports_set_status(resource, event_type,
                                                   mock.ANY, fake_payload)
            mock_set_subport_status.assert_called_once_with(
                core_plugin, mock.ANY, 'fake_port_id',
                n_const.PORT_STATUS_ACTIVE)

    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_set_subport_status')
    def test_trunk_subports_set_status_create_parent_down(
            self, mock_set_subport_status):
        resource = resources.SUBPORTS
        event_type = events.AFTER_CREATE
        fake_payload = self._fake_trunk_payload()
        core_plugin = directory.get_plugin()

        fake_payload.subports = [models.SubPort(port_id='fake_port_id',
                                                segmentation_id=101,
                                                segmentation_type='vlan',
                                                trunk_id='fake_id')]
        parent_port = FAKE_PARENT.copy()
        parent_port['status'] = n_const.PORT_STATUS_DOWN

        with mock.patch.object(core_plugin, '_get_port') as gp:
            gp.return_value = parent_port
            self.handler.trunk_subports_set_status(resource, event_type,
                                                   mock.ANY, fake_payload)
            mock_set_subport_status.assert_called_once_with(
                core_plugin, mock.ANY, 'fake_port_id',
                n_const.PORT_STATUS_DOWN)

    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_set_subport_status')
    def test_trunk_subports_set_status_delete(self, mock_set_subport_status):
        resource = resources.SUBPORTS
        event_type = events.AFTER_DELETE
        fake_payload = self._fake_trunk_payload()

        fake_payload.subports = [models.SubPort(port_id='fake_port_id',
                                                segmentation_id=101,
                                                segmentation_type='vlan',
                                                trunk_id='fake_id')]

        self.handler.trunk_subports_set_status(resource, event_type, mock.ANY,
                                               fake_payload)
        mock_set_subport_status.assert_called_once_with(
            mock.ANY, mock.ANY, 'fake_port_id', n_const.PORT_STATUS_DOWN)

    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_get_subports_ids')
    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_set_subport_status')
    def test_trunk_subports_update_status_parent_down_to_active(
            self, mock_set_subport_status, mock_get_subports_ids):
        resource = resources.PORT
        event_type = events.AFTER_UPDATE
        core_plugin = directory.get_plugin()
        port = FAKE_PARENT.copy()
        original_port = FAKE_PARENT.copy()
        original_port['status'] = n_const.PORT_STATUS_DOWN
        payload = events.DBEventPayload(None, None, None,
                                        states=(original_port, port))

        mock_get_subports_ids.return_value = ['fake_port_id']

        self.handler.trunk_subports_update_status(resource, event_type,
                                                  mock.ANY, payload)

        mock_set_subport_status.assert_called_once_with(
            core_plugin, mock.ANY, 'fake_port_id', n_const.PORT_STATUS_ACTIVE)

    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_get_subports_ids')
    @mock.patch.object(trunk_driver.OpenDaylightTrunkHandlerV2,
                       '_set_subport_status')
    def test_trunk_subports_update_status_parent_active_to_down(
            self, mock_set_subport_status, mock_get_subports_ids):
        resource = resources.PORT
        event_type = events.AFTER_UPDATE
        core_plugin = directory.get_plugin()
        port = FAKE_PARENT.copy()
        original_port = FAKE_PARENT.copy()
        port['status'] = n_const.PORT_STATUS_DOWN
        payload = events.DBEventPayload(None, None, None,
                                        states=(original_port, port))

        mock_get_subports_ids.return_value = ['fake_port_id']

        self.handler.trunk_subports_update_status(resource, event_type,
                                                  mock.ANY, payload)

        mock_set_subport_status.assert_called_once_with(
            core_plugin, mock.ANY, 'fake_port_id', n_const.PORT_STATUS_DOWN)


class TestTrunkDriver(base_v2.OpenDaylightConfigBase):
    def setUp(self):
        super(TestTrunkDriver, self).setUp()

    def test_is_loaded(self):
        driver = trunk_driver.OpenDaylightTrunkDriverV2.create()
        self.cfg.config(mechanism_drivers=["logger",
                                           odl_const.ODL_ML2_MECH_DRIVER_V2],
                        group='ml2')
        self.assertTrue(driver.is_loaded)

        self.cfg.config(mechanism_drivers=['logger'], group='ml2')
        self.assertFalse(driver.is_loaded)

        self.cfg.config(core_plugin='some_plugin')
        self.assertFalse(driver.is_loaded)



# Copyright (c) 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import atexit
import os

from neutron_lib import worker
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall

# TODO(mpeterson): this should go back to the previous block once the problems
# with flake8-import-order are fixed.
from neutron.agent.linux import daemon

from networking_odl._i18n import _
from networking_odl.journal import cleanup
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.journal import periodic_task
from networking_odl.journal import recovery


LOG = logging.getLogger(__name__)


class JournalPeriodicProcessor(worker.BaseWorker):
    """Responsible for running the periodic processing of the journal.

    This is a separate worker as the regular journal thread is called when an
    operation finishes and that run will take care of any and all entries
    that might be present in the journal, including the one relating to that
    operation.

    A periodic run over the journal is thus necessary for cases when journal
    entries in the aforementioned run didn't process correctly due to some
    error (usually a connection problem) and need to be retried.
    """
    def __init__(self):
        super(JournalPeriodicProcessor, self).__init__()
        self._journal = journal.OpenDaylightJournalThread(start_thread=False)
        self._interval = cfg.CONF.ml2_odl.sync_timeout
        self._timer = None
        self._maintenance_task = None
        self._running = None
        self.pidfile = None

    def _create_pidfile(self):
        pidfile = os.path.join(cfg.CONF.state_path,
                               type(self).__name__.lower() + '.pid')
        self.pidfile = daemon.Pidfile(pidfile, 'python')

        # NOTE(mpeterson): We want self._running to be None before the first
        # run so atexit is only registered once and not several times.
        if self._running is None:
            atexit.unregister(self._delete_pidfile)
            atexit.register(self._delete_pidfile)

        self.pidfile.write(os.getpid())

    def _delete_pidfile(self):
        if self.pidfile is not None:
            self.pidfile.unlock()
            os.remove(str(self.pidfile))
            self.pidfile = None

    def start(self):
        if self._running:
            raise RuntimeError(
                _("Thread has to be stopped before started again")
            )

        super(JournalPeriodicProcessor, self).start()
        LOG.debug('JournalPeriodicProcessor starting')
        self._journal.start()
        self._timer = loopingcall.FixedIntervalLoopingCall(self._call_journal)
        self._timer.start(self._interval)
        self._start_maintenance_task()
        self._create_pidfile()
        self._running = True

    def stop(self):
        if not self._running:
            return

        LOG.debug('JournalPeriodicProcessor stopping')
        self._journal.stop()
        self._timer.stop()
        self._maintenance_task.cleanup()
        self._delete_pidfile()
        super(JournalPeriodicProcessor, self).stop()
        self._running = False

    def wait(self):
        pass

    def reset(self):
        if self._maintenance_task is not None:
            self._maintenance_task.execute_ops(forced=True)

    def _call_journal(self):
        self._journal.set_sync_event()

    def _start_maintenance_task(self):
        self._maintenance_task = periodic_task.PeriodicTask(
            'maintenance', cfg.CONF.ml2_odl.maintenance_interval)

        for phase in (
                cleanup.delete_completed_rows,
                cleanup.cleanup_processing_rows,
                full_sync.full_sync,
                recovery.journal_recovery,
        ):
            self._maintenance_task.register_operation(phase)

        self._maintenance_task.execute_ops(forced=True)
        self._maintenance_task.start()



# Copyright (c) 2015 Intel Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests import base
from oslo_config import fixture as config_fixture

from networking_odl.common import lightweight_testing as lwt
from networking_odl.tests import base as odl_base


class LightweightTestingTestCase(base.DietTestCase):
    def setUp(self):
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        self.cfg = self.useFixture(config_fixture.Config())
        super(LightweightTestingTestCase, self).setUp()

    def test_create_client_with_lwt_enabled(self):
        """Have to do the importation here, otherwise there will be a loop"""
        from networking_odl.common import client as odl_client
        self.cfg.config(enable_lightweight_testing=True, group='ml2_odl')
        # DietTestCase does not automatically cleans configuration overrides
        self.addCleanup(odl_client.cfg.CONF.reset)

        client = odl_client.OpenDaylightRestClient.create_client()
        self.assertIsInstance(client, lwt.OpenDaylightLwtClient)

    def test_create_client_with_lwt_disabled(self):
        """Have to do the importation here, otherwise there will be a loop"""
        from networking_odl.common import client as odl_client
        self.cfg.config(enable_lightweight_testing=False, group='ml2_odl')
        # DietTestCase does not automatically cleans configuration overrides
        self.addCleanup(odl_client.cfg.CONF.reset)

        client = odl_client.OpenDaylightRestClient.create_client()
        self.assertIsInstance(client, odl_client.OpenDaylightRestClient)

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'networks': {}}, clear=True)
    def test_post_single_resource(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        fake_network1 = {'id': 'fakeid1', 'name': 'fake_network1'}
        obj = {'networks': fake_network1}
        response = client.sendjson('post', 'networks', obj)
        self.assertEqual(lwt.NO_CONTENT, response.status_code)
        lwt_dict = lwt.OpenDaylightLwtClient.lwt_dict
        self.assertEqual(lwt_dict['networks']['fakeid1'],
                         fake_network1)

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'networks': {}}, clear=True)
    def test_post_multiple_resources(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        fake_network1 = {'id': 'fakeid1', 'name': 'fake_network1'}
        fake_network2 = {'id': 'fakeid2', 'name': 'fake_network2'}
        obj = {'networks': [fake_network1, fake_network2]}
        response = client.sendjson('post', 'networks', obj)
        self.assertEqual(lwt.NO_CONTENT, response.status_code)
        lwt_dict = lwt.OpenDaylightLwtClient.lwt_dict
        self.assertEqual(lwt_dict['networks']['fakeid1'],
                         fake_network1)
        self.assertEqual(lwt_dict['networks']['fakeid2'],
                         fake_network2)

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'ports': {'fakeid1': {'id': 'fakeid1',
                                            'name': 'fake_port1'}}},
                     clear=True)
    def test_get_single_resource(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        url_path = 'ports/fakeid1'
        response = client.sendjson('get', url_path, None)
        self.assertEqual(lwt.OK, response.status_code)
        res = response.json()
        # For single resource, the return value is a dict
        self.assertEqual(res['port']['name'], 'fake_port1')

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'ports': {'fakeid1': {'id': 'fakeid1',
                                            'name': 'fake_port1'},
                                'fakeid2': {'id': 'fakeid2',
                                            'name': 'fake_port2'}}},
                     clear=True)
    def test_get_multiple_resources(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        url_path = 'ports/'
        response = client.sendjson('get', url_path, None)
        self.assertEqual(lwt.OK, response.status_code)
        res = response.json()
        for port in res:
            self.assertIn(port['port']['name'],
                          ['fake_port1', 'fake_port2'])

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'subnets': {'fakeid1': {'id': 'fakeid1',
                                              'name': 'fake_subnet1'}}},
                     clear=True)
    def test_put_single_resource(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        changed = {'id': 'fakeid1', 'name': 'fake_subnet1_changed'}
        obj = {'subnets': changed}

        url_path = 'subnets/fakeid1'
        response = client.sendjson('put', url_path, obj)
        self.assertEqual(lwt.NO_CONTENT, response.status_code)
        lwt_dict = lwt.OpenDaylightLwtClient.lwt_dict
        self.assertEqual('fake_subnet1_changed',
                         lwt_dict['subnets']['fakeid1']['name'])

        """Check the client does not change the parameter"""
        self.assertEqual('fakeid1', changed['id'])
        self.assertEqual('fake_subnet1_changed', changed['name'])

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'subnets': {'fakeid1': {'id': 'fakeid1',
                                              'name': 'fake_subnet1'},
                                  'fakeid2': {'id': 'fakeid2',
                                              'name': 'fake_subnet2'}}},
                     clear=True)
    def test_put_multiple_resources(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        changed1 = {'id': 'fakeid1', 'name': 'fake_subnet1_changed'}
        changed2 = {'id': 'fakeid2', 'name': 'fake_subnet2_changed'}
        obj = {'subnets': [changed1, changed2]}

        url_path = 'subnets/'
        response = client.sendjson('put', url_path, obj)
        self.assertEqual(lwt.NO_CONTENT, response.status_code)
        lwt_dict = lwt.OpenDaylightLwtClient.lwt_dict
        self.assertEqual('fake_subnet1_changed',
                         lwt_dict['subnets']['fakeid1']['name'])
        self.assertEqual('fake_subnet2_changed',
                         lwt_dict['subnets']['fakeid2']['name'])

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'networks': {'fakeid1': {'id': 'fakeid1',
                                               'name': 'fake_network1'}}},
                     clear=True)
    def test_delete_single_resource(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        url_path = 'networks/fakeid1'
        response = client.sendjson('delete', url_path, None)
        self.assertEqual(lwt.NO_CONTENT, response.status_code)
        lwt_dict = lwt.OpenDaylightLwtClient.lwt_dict
        network = lwt_dict['networks'].get('fakeid1')
        self.assertIsNone(network)

    @mock.patch.dict(lwt.OpenDaylightLwtClient.lwt_dict,
                     {'networks': {'fakeid1': {'id': 'fakeid1',
                                               'name': 'fake_network1'},
                                   'fakeid2': {'id': 'fakeid2',
                                               'name': 'fake_network2'}}},
                     clear=True)
    def test_delete_multiple_resources(self):
        client = lwt.OpenDaylightLwtClient.create_client()
        network1 = {'id': 'fakeid1'}
        network2 = {'id': 'fakeid2'}
        obj = {'networks': [network1, network2]}
        response = client.sendjson('delete', 'networks/', obj)
        self.assertEqual(lwt.NO_CONTENT, response.status_code)
        lwt_dict = lwt.OpenDaylightLwtClient.lwt_dict
        network = lwt_dict['networks'].get('fakeid1')
        self.assertIsNone(network)
        network = lwt_dict['networks'].get('fakeid2')
        self.assertIsNone(network)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib import constants as n_const
from neutron_lib.plugins import utils as p_utils
from oslo_log import log as logging

from networking_odl.common import constants

LOG = logging.getLogger(__name__)


class OdlDhcpDriverBase(object):

    # NOTE:(Karthik Prasad/karthik.prasad) Not validating based on value change
    # of enable_dhcp in case of subnet update event, instead validating on
    # port_id presence in DB by locking the session, this will enable user to
    # reissue the same command in case of failure.
    def create_or_delete_dhcp_port(self, subnet_context):
        # NOTE:(Achuth) Fixes bug 1746715
        # DHCP port to be created for IPv4 subnets only, since ODL doesn't
        # support IPv6 neutron port ARP responses. This prevents validations
        # in ODL and avoids processing  these ports incorrectly.
        if subnet_context.current['ip_version'] != 4:
            LOG.warning("ODL DHCP port is supported  only for IPv4 subnet %s",
                        subnet_context.current['id'])
            return
        port_id = self.get_dhcp_port_if_exists(subnet_context)
        plugin = subnet_context._plugin
        if not port_id and subnet_context.current['enable_dhcp']:
            LOG.debug("Creating ODL DHCP port for subnet %s of network %s",
                      subnet_context.current['id'],
                      subnet_context.current['network_id'])
            port = self._make_dhcp_port_dict(subnet_context)
            # TODO(boden): rehome and consume from neutron-lib
            p_utils.create_port(plugin, subnet_context._plugin_context, port)
        if port_id and not subnet_context.current['enable_dhcp']:
            self._delete_port(plugin, subnet_context._plugin_context, port_id)

    def _delete_port(self, plugin, context, port_id):
        LOG.debug("Deleting ODL DHCP port with id %s", port_id)
        plugin.delete_port(context, port_id)

    def _make_dhcp_port_dict(self, subnet_context):

        subnet_id = subnet_context.current['id']
        port_dict = dict(
            name='',
            admin_state_up=True,
            device_id=constants.ODL_DEVICE_ID_START + '-' + subnet_id,
            device_owner=n_const.DEVICE_OWNER_DHCP,
            network_id=subnet_context.current['network_id'],
            fixed_ips=[dict(subnet_id=subnet_id)],
            tenant_id=subnet_context.network.current['tenant_id'])

        return {'port': port_dict}

    def get_dhcp_port_if_exists(self, subnet_context):

        plugin = subnet_context._plugin
        plugin_context = subnet_context._plugin_context
        network_id = subnet_context._subnet['network_id']
        subnet_id = subnet_context.current['id']
        device_id = constants.ODL_DEVICE_ID_START + '-' + subnet_id
        LOG.debug("Retrieving ODL DHCP port for subnet %s", subnet_id)
        filters = {
            'network_id': [network_id],
            'device_id': [device_id],
            'device_owner': [n_const.DEVICE_OWNER_DHCP]
        }
        ports = plugin.get_ports(plugin_context, filters=filters)
        if ports:
            port = ports[0]
            LOG.debug("Retrieved ODL owned port %s for subnet %s",
                      port['id'], subnet_id)
            return port['id']
        return None



# Copyright (C) 2017 Intel Corporation.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import contextlib

from oslo_utils import uuidutils

from neutron.extensions import qos as qos_ext
from neutron.services.qos import qos_plugin
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib import fixture as nlib_fixture
from neutron_lib.plugins import directory

from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base


class QoSTestExtensionManager(object):

    def get_resources(self):
        return qos_ext.Qos.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class _QoSDriverTestCase(base.OdlTestsBase):

    def test_policy_create(self):
        with self.qos_policy() as policy:
            self.assert_resource_created(
                odl_const.ODL_QOS_POLICY, policy)

    def test_policy_update(self):
        with self.qos_policy() as policy:
            self.resource_update_test(
                odl_const.ODL_QOS_POLICY, policy)

    def test_policy_delete(self):
        with self.qos_policy() as policy:
            self.resource_delete_test(
                odl_const.ODL_QOS_POLICY, policy)


class QoSDriverTests(base.V2DriverAdjustment,
                     _QoSDriverTestCase,
                     test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']

    def setUp(self):
        self.useFixture(nlib_fixture.PluginDirectoryFixture())
        super(QoSDriverTests, self).setUp()
        self.qos_plug = qos_plugin.QoSPlugin()
        directory.add_plugin('QOS', self.qos_plug)
        ext_mgr = QoSTestExtensionManager()
        self.resource_prefix_map = {'policies': '/qos'}
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        tenant_id = uuidutils.generate_uuid()
        self.policy_data = {
            'policy': {'name': 'test-policy', 'tenant_id': tenant_id}}

    @contextlib.contextmanager
    def qos_policy(self, fmt='json'):
        po_res = self.new_create_request('policies', self.policy_data, fmt)
        po_rep = po_res.get_response(self.ext_api)
        policy = self.deserialize(fmt, po_rep)
        yield policy



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from datetime import timedelta

from neutron_lib.db import api as db_api
from oslo_config import cfg
from oslo_log import log as logging

from networking_odl.common import constants as odl_const
from networking_odl.db import db

LOG = logging.getLogger(__name__)


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def delete_completed_rows(context):
    """Journal maintenance operation for deleting completed rows."""
    rows_retention = cfg.CONF.ml2_odl.completed_rows_retention
    if rows_retention <= 0:
        return

    LOG.debug("Deleting completed rows")
    db.delete_rows_by_state_and_time(
        context, odl_const.COMPLETED,
        timedelta(seconds=rows_retention))


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def cleanup_processing_rows(context):
    row_count = db.reset_processing_rows(
        context, cfg.CONF.ml2_odl.processing_timeout)
    if row_count:
        LOG.info("Reset %(num)s orphaned rows back to pending",
                 {"num": row_count})
